from django.core.management.base import BaseCommand

from course_flow.models import Activity, Column, Node, NodeWeek, Week


class Command(BaseCommand):
    def add_arguments(self, parser):
        pass

    def handle(self, *args, **kwargs):
        # JIGSAW
        try:
            workflow = Activity.objects.get(
                title="Jigsaw",
                is_strategy=True,
                published=True,
                from_saltise=True,
            )
            workflow.weeks.first().nodes.all().delete()
        except:
            workflow = Activity.objects.create(
                title="Jigsaw",
                is_strategy=True,
                published=True,
                from_saltise=True,
            )
        week = workflow.weeks.first()
        week.strategy_classification = Week.JIGSAW
        week.save()
        rank = 0
        ooci = workflow.columns.get(column_type=Column.OUT_OF_CLASS_INSTRUCTOR)
        oocs = workflow.columns.get(column_type=Column.OUT_OF_CLASS_STUDENT)
        ici = workflow.columns.get(column_type=Column.IN_CLASS_INSTRUCTOR)
        ics = workflow.columns.get(column_type=Column.IN_CLASS_STUDENT)
        node = week.nodes.create(
            title="Divide students into groups",
            description="Instructor divides students into small groups of five or six students.",
            column=ici,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Assign a portion of an assignment",
            description="Instructor prepares an assignment and divides it into as many parts as there are students in each group. Each member of the group is assigned a portion of the assignment or research project to complete.",
            column=ici,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Research Material",
            description="In groups, students research the material(s) pertaining to their section of the assignment and prepare to discuss it with their classmates.",
            column=ics,
            context_classification=Node.GROUPS,
            task_classification=Node.GATHER_INFO,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Expert Groups",
            description='Students form temporary "expert groups" with others who have been assigned the same portion of the assignment and discuss the material they have covered, and prepare to present this material to their original "jigsaw" group.',
            column=ics,
            context_classification=Node.GROUPS,
            task_classification=Node.DISCUSS,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Present to original group",
            description='Students return to their original "jigsaw" group and present the material they have covered.',
            column=ics,
            context_classification=Node.INDIVIDUAL,
            task_classification=Node.PRESENT,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Complete assignment",
            description="Using their collective knowledge, students complete their assignment.",
            column=ics,
            context_classification=Node.GROUPS,
            task_classification=Node.WRITE,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Evaluation",
            description="Instructor evaluates the completed assignment. Option: Additional assessment of the students' knowledge (individual or group) can be added.",
            column=ici,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        for nodeweek in week.nodeweek_set.all():
            print(nodeweek.rank)

        # PEER INSTRUCTION
        try:
            workflow = Activity.objects.get(
                title="Peer Instruction",
                is_strategy=True,
                published=True,
                from_saltise=True,
            )
            workflow.weeks.first().nodes.all().delete()
        except:
            workflow = Activity.objects.create(
                title="Peer Instruction",
                is_strategy=True,
                published=True,
                from_saltise=True,
            )
        week = workflow.weeks.first()
        week.strategy_classification = Week.PEER_INSTRUCTION
        week.save()
        rank = 0
        ooci = workflow.columns.get(column_type=Column.OUT_OF_CLASS_INSTRUCTOR)
        oocs = workflow.columns.get(column_type=Column.OUT_OF_CLASS_STUDENT)
        ici = workflow.columns.get(column_type=Column.IN_CLASS_INSTRUCTOR)
        ics = workflow.columns.get(column_type=Column.IN_CLASS_STUDENT)
        node = week.nodes.create(
            title="Initial Content",
            description="Instructor provides students with content (e.g., mini-lecture, pre-class reading).",
            column=ici,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Multiple Choice",
            description="Instructor prepares multiple choice questions/problems.",
            column=ici,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Individual answers",
            description="Individually, students answer the question and enter their vote using a polling method or device (e.g., clickers or phone app).",
            column=ics,
            context_classification=Node.INDIVIDUAL,
            task_classification=Node.PROBLEM_SOLVE,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Group Discussion",
            description="Students, in pairs or small groups, discuss their answers and explain their reasoning.",
            column=ics,
            context_classification=Node.GROUPS,
            task_classification=Node.DISCUSS,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Re-answer",
            description="Individually, students vote again and select answer based on discussion - they can either keep their original answer or change.",
            column=ics,
            context_classification=Node.INDIVIDUAL,
            task_classification=Node.PROBLEM_SOLVE,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Review",
            description="Instructor briefly reviews the concepts.",
            column=ici,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1

        for nodeweek in week.nodeweek_set.all():
            print(nodeweek.rank)

        # CASE STUDIES
        try:
            workflow = Activity.objects.get(
                title="Case Studies",
                is_strategy=True,
                published=True,
                from_saltise=True,
            )
            workflow.weeks.first().nodes.all().delete()
        except:
            workflow = Activity.objects.create(
                title="Case Studies",
                is_strategy=True,
                published=True,
                from_saltise=True,
            )
        week = workflow.weeks.first()
        week.strategy_classification = Week.CASE_STUDIES
        week.save()
        rank = 0
        ooci = workflow.columns.get(column_type=Column.OUT_OF_CLASS_INSTRUCTOR)
        oocs = workflow.columns.get(column_type=Column.OUT_OF_CLASS_STUDENT)
        ici = workflow.columns.get(column_type=Column.IN_CLASS_INSTRUCTOR)
        ics = workflow.columns.get(column_type=Column.IN_CLASS_STUDENT)
        node = week.nodes.create(
            title="Distrubute Case",
            description="Instructor selects and distributes case study (or studies) that they want students to work with.",
            column=ici,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Read and prepare",
            description="Students - individually or in groups - read the case study and identify key components.",
            context_classification=Node.INDIVIDUAL,
            task_classification=Node.READ,
            column=ics,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Discuss",
            description="In groups, students discuss key components and lessons learned - i.e., general principles or exceptions.",
            column=ics,
            context_classification=Node.GROUPS,
            task_classification=Node.DISCUSS,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Distribute new case",
            description="Instructor distributes a new case.",
            column=ici,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Solve new case",
            description="In groups, students work with the new case, applying lessons learned to solve it. Students apply general principles and reason as to whether or not the new case is typical or an exception.",
            column=ics,
            context_classification=Node.GROUPS,
            task_classification=Node.DISCUSS,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Class discussion",
            description="As a class, students discuss key components and lessons learned - i.e. general principles or exceptions.",
            column=ics,
            context_classification=Node.WHOLE_CLASS,
            task_classification=Node.DISCUSS,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1

        for nodeweek in week.nodeweek_set.all():
            print(nodeweek.rank)

        # GALLERY WALK
        try:
            workflow = Activity.objects.get(
                title="Gallery Walk",
                is_strategy=True,
                published=True,
                from_saltise=True,
            )
            workflow.weeks.first().nodes.all().delete()
        except:
            workflow = Activity.objects.create(
                title="Gallery Walk",
                is_strategy=True,
                published=True,
                from_saltise=True,
            )
        week = workflow.weeks.first()
        week.strategy_classification = Week.GALLERY_WALK
        week.save()
        rank = 0
        ooci = workflow.columns.get(column_type=Column.OUT_OF_CLASS_INSTRUCTOR)
        oocs = workflow.columns.get(column_type=Column.OUT_OF_CLASS_STUDENT)
        ici = workflow.columns.get(column_type=Column.IN_CLASS_INSTRUCTOR)
        ics = workflow.columns.get(column_type=Column.IN_CLASS_STUDENT)
        node = week.nodes.create(
            title="Assign problems",
            description="Instructor assigns a problem to each group of students. Problems may be identical or distinct, but they generally address related content.",
            column=ici,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Solve problems",
            description="In groups, students try to solve their assigned problem.",
            context_classification=Node.GROUPS,
            task_classification=Node.PROBLEM_SOLVE,
            column=ics,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Peer review",
            description="In groups, students move around the classroom and peer-review the work completed by another group, making annotations and providing feedback.",
            column=ics,
            context_classification=Node.GROUPS,
            task_classification=Node.ASSESS_PEERS,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Analyze feedback",
            description="In groups, students examine the feedback provided by their peers before finalizing their response.",
            column=ics,
            context_classification=Node.GROUPS,
            task_classification=Node.DISCUSS,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Class discussion",
            description="Instructor evaluates all responses with the class, highlighting any errors and distinct procedures for arriving at correct solutions.",
            column=ici,
            context_classification=Node.WHOLE_CLASS,
            task_classification=Node.DISCUSS,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1

        for nodeweek in week.nodeweek_set.all():
            print(nodeweek.rank)

        # REFLECTIVE WRITING
        try:
            workflow = Activity.objects.get(
                title="Reflective Writing",
                is_strategy=True,
                published=True,
                from_saltise=True,
            )
            workflow.weeks.first().nodes.all().delete()
        except:
            workflow = Activity.objects.create(
                title="Reflective Writing",
                is_strategy=True,
                published=True,
                from_saltise=True,
            )
        week = workflow.weeks.first()
        week.strategy_classification = Week.REFLECTIVE_WRITING
        week.save()
        rank = 0
        ooci = workflow.columns.get(column_type=Column.OUT_OF_CLASS_INSTRUCTOR)
        oocs = workflow.columns.get(column_type=Column.OUT_OF_CLASS_STUDENT)
        ici = workflow.columns.get(column_type=Column.IN_CLASS_INSTRUCTOR)
        ics = workflow.columns.get(column_type=Column.IN_CLASS_STUDENT)
        node = week.nodes.create(
            title="Assign reading",
            description="Instructor assigns reading or viewing material(s) - a section of textbook, video, etc. Specific prompts are provided to guide this task, which can include what is not understood or unclear and why it is unclear, or what is understood and how it might related to what the student already knows or to other ideas from the course.",
            column=ooci,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Read material",
            description="Individually, students review the material with the intention of documenting in writing their understanding using the prompts.",
            column=oocs,
            context_classification=Node.INDIVIDUAL,
            task_classification=Node.READ,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Reflective writing",
            description="Students reflect on the material, identifying the items they understand and those with which they are struggling. They then formulate short paragraphs in response to the prompts and submit their writings to the instructor.",
            column=oocs,
            context_classification=Node.INDIVIDUAL,
            task_classification=Node.WRITE,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Evaluation and analysis",
            description="Instructor reviews students' work and uses it to determine the lesson plan, such as which topics need further review and explanation or what activities or materials can be used next.",
            column=ooci,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1

        for nodeweek in week.nodeweek_set.all():
            print(nodeweek.rank)

        # TWO STAGE EXAM
        try:
            workflow = Activity.objects.get(
                title="Two Stage Exam",
                is_strategy=True,
                published=True,
                from_saltise=True,
            )
            workflow.weeks.first().nodes.all().delete()
        except:
            workflow = Activity.objects.create(
                title="Two Stage Exam",
                is_strategy=True,
                published=True,
                from_saltise=True,
            )
        week = workflow.weeks.first()
        week.strategy_classification = Week.TWO_STAGE_EXAM
        week.save()
        rank = 0
        ooci = workflow.columns.get(column_type=Column.OUT_OF_CLASS_INSTRUCTOR)
        oocs = workflow.columns.get(column_type=Column.OUT_OF_CLASS_STUDENT)
        ici = workflow.columns.get(column_type=Column.IN_CLASS_INSTRUCTOR)
        ics = workflow.columns.get(column_type=Column.IN_CLASS_STUDENT)
        node = week.nodes.create(
            title="Preparation",
            description="Prior to class, instructor prepares two levels of assessment - individual and group, then in class distributes the individual assessment.",
            column=ici,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Individual assessment",
            description="Individually, students complete the assessment.",
            context_classification=Node.INDIVIDUAL,
            task_classification=Node.PROBLEM_SOLVE,
            column=ics,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Assign groups",
            description="Instructor collects the assessment then splits the students into small groups (3-4 students).",
            column=ici,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Group Discussion",
            description="As groups, students discuss their answers then work together to answer the new questions.",
            column=ics,
            context_classification=Node.GROUPS,
            task_classification=Node.DISCUSS,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Group assessment",
            description="As a group, students complete the group assessment.",
            column=ics,
            context_classification=Node.GROUPS,
            task_classification=Node.PROBLEM_SOLVE,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Evaluation",
            description="Instructor collects the group assessment, and evaluates both.",
            column=ici,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1

        for nodeweek in week.nodeweek_set.all():
            print(nodeweek.rank)

        # TOOLKIT
        try:
            workflow = Activity.objects.get(
                title="Toolkit",
                is_strategy=True,
                published=True,
                from_saltise=True,
            )
            workflow.weeks.first().nodes.all().delete()
        except:
            workflow = Activity.objects.create(
                title="Toolkit",
                is_strategy=True,
                published=True,
                from_saltise=True,
            )
        week = workflow.weeks.first()
        week.strategy_classification = Week.TOOLKIT
        week.save()
        rank = 0
        ooci = workflow.columns.get(column_type=Column.OUT_OF_CLASS_INSTRUCTOR)
        oocs = workflow.columns.get(column_type=Column.OUT_OF_CLASS_STUDENT)
        ici = workflow.columns.get(column_type=Column.IN_CLASS_INSTRUCTOR)
        ics = workflow.columns.get(column_type=Column.IN_CLASS_STUDENT)
        node = week.nodes.create(
            title="Give students a template",
            description="Instructor provides students, at the end of a series of lessons/activities, with a template to support their identification and consolidation of the major intellectual tools learned and used in the preceding lessons.",
            column=ici,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Discussion of themes or principles",
            description="In groups, students discuss the lesson content and identify use-case examples of discipline-relevant principles, themes, rules/algorithms, etc.",
            context_classification=Node.GROUPS,
            task_classification=Node.DISCUSS,
            column=ics,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Presentation of toolkit",
            description="In groups, students present their list to the whole class.",
            context_classification=Node.GROUPS,
            task_classification=Node.PRESENT,
            column=ics,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Feedback",
            description="Instructor provides feedback on the relevance of the items to be included in the Toolkit list and guides students in their review, evaluation, and decision-making.",
            column=ici,
            context_classification=Node.WHOLE_CLASS,
            task_classification=Node.DISCUSS,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Discuss and consolidate",
            description="As a class, students discuss, evaluate, and justify decisions for entry of items into a single list.",
            column=ics,
            context_classification=Node.WHOLE_CLASS,
            task_classification=Node.DISCUSS,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1

        for nodeweek in week.nodeweek_set.all():
            print(nodeweek.rank)

        # ONE MINUTE PAPER
        try:
            workflow = Activity.objects.get(
                title="One Minute Paper",
                is_strategy=True,
                published=True,
                from_saltise=True,
            )
            workflow.weeks.first().nodes.all().delete()
        except:
            workflow = Activity.objects.create(
                title="One Minute Paper",
                is_strategy=True,
                published=True,
                from_saltise=True,
            )
        week = workflow.weeks.first()
        week.strategy_classification = Week.ONE_MINUTE_PAPER
        week.save()
        rank = 0
        ooci = workflow.columns.get(column_type=Column.OUT_OF_CLASS_INSTRUCTOR)
        oocs = workflow.columns.get(column_type=Column.OUT_OF_CLASS_STUDENT)
        ici = workflow.columns.get(column_type=Column.IN_CLASS_INSTRUCTOR)
        ics = workflow.columns.get(column_type=Column.IN_CLASS_STUDENT)
        node = week.nodes.create(
            title="Assign reflection",
            description="Instructor asks students to write a brief reflection on their understanding of a lesson or activity. A time limit is provided (1 min or more).",
            column=ici,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Discussion of themes or principles",
            description="In groups, students discuss the lesson content and identify use-case examples of discipline-relevant principles, themes, rules/algorithms, etc.",
            context_classification=Node.GROUPS,
            task_classification=Node.DISCUSS,
            column=ics,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="One minute response",
            description="Individually, students provide written responses, addressing such issues as the most important thing learned during this class or what important questions remain unanswered.",
            context_classification=Node.INDIVIDUAL,
            task_classification=Node.WRITE,
            column=ics,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Analysis of responses",
            description="Instructor collects written reflections and uses them to determine the lesson plan for the next class, including which topics need further review or explanation and what activities or material can be used next.",
            column=ici,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1

        for nodeweek in week.nodeweek_set.all():
            print(nodeweek.rank)

        # DISTRIBUTED PROBLEM SOLVING
        try:
            workflow = Activity.objects.get(
                title="Distributed Problem Solving",
                is_strategy=True,
                published=True,
                from_saltise=True,
            )
            workflow.weeks.first().nodes.all().delete()
        except:
            workflow = Activity.objects.create(
                title="Distributed Problem Solving",
                is_strategy=True,
                published=True,
                from_saltise=True,
            )
        week = workflow.weeks.first()
        week.strategy_classification = Week.DISTRIBUTED_PROBLEM_SOLVING
        week.save()
        rank = 0
        ooci = workflow.columns.get(column_type=Column.OUT_OF_CLASS_INSTRUCTOR)
        oocs = workflow.columns.get(column_type=Column.OUT_OF_CLASS_STUDENT)
        ici = workflow.columns.get(column_type=Column.IN_CLASS_INSTRUCTOR)
        ics = workflow.columns.get(column_type=Column.IN_CLASS_STUDENT)
        node = week.nodes.create(
            title="Assign groups",
            description="Instructor divides the class into small groups (4-6 students) and assigns each to a shared writable surface (whiteboard or interactive board).",
            column=ici,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Assign distinct problems",
            description="Instructor assigns one problem for each group. Problems should be closely related but distinct, and there should be at least two different problems.",
            column=ici,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Begin problem solving",
            description="Groups begin their problem solving process, making sure to record their worked solutions on the surface, e.g. Group 1 working on Problem A, Group 2 working on Problem B, etc.",
            column=ics,
            context_classification=Node.GROUPS,
            task_classification=Node.PROBLEM_SOLVE,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Interrupt process and switch groups",
            description="Instructor interrupts the process and has students switch problems and take up another group's problem-solving process, e.g. Group 1 moves to Group 2's problem B, Group 2 moves to Group 1's Problem A.",
            column=ici,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Review work of other group",
            description="Groups review and revise the solution of the first group and decide how they will continue the problem solving process.",
            column=ics,
            context_classification=Node.GROUPS,
            task_classification=Node.ASSESS_PEERS,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Complete Problem",
            description="Groups complete the other group's problem.",
            column=ics,
            context_classification=Node.GROUPS,
            task_classification=Node.PROBLEM_SOLVE,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Interrupt process again",
            description="Instructor interrupts the process and instructs groups to return to their original problem and evaluate the completed solution.",
            column=ici,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Evaluate solution to original problem",
            description="Groups review and evaluate the final solution to the original problem - e.g. Group 1 returns to Problem A, which was completed by Group 2.",
            column=ics,
            context_classification=Node.GROUPS,
            task_classification=Node.ASSESS_PEERS,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Instructor evaluation and discussion",
            description="Instructor evaluates all responses with the class, highlighting any errors and distinct procedures for arriving at correct solutions.",
            column=ici,
            context_classification=Node.WHOLE_CLASS,
            task_classification=Node.DISCUSS,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        for nodeweek in week.nodeweek_set.all():
            print(nodeweek.rank)

        # PEER ASSESSMENT
        try:
            workflow = Activity.objects.get(
                title="Peer Assessment",
                is_strategy=True,
                published=True,
                from_saltise=True,
            )
            workflow.weeks.first().nodes.all().delete()
        except:
            workflow = Activity.objects.create(
                title="Peer Assessment",
                is_strategy=True,
                published=True,
                from_saltise=True,
            )
        week = workflow.weeks.first()
        week.strategy_classification = Week.REFLECTIVE_WRITING
        week.save()
        rank = 0
        ooci = workflow.columns.get(column_type=Column.OUT_OF_CLASS_INSTRUCTOR)
        oocs = workflow.columns.get(column_type=Column.OUT_OF_CLASS_STUDENT)
        ici = workflow.columns.get(column_type=Column.IN_CLASS_INSTRUCTOR)
        ics = workflow.columns.get(column_type=Column.IN_CLASS_STUDENT)
        node = week.nodes.create(
            title="Assign rubric to students for peer assessment",
            description="Instructor assigns students, individually or in groups, to review the work produced by peers using a rubric or template.",
            column=ici,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Provide feedback using rubric",
            description="Individually or in small groups, students review another individual's or group's work using the rubric or template. Students provide peers with feedback in writing or prepared for oral presentation.",
            column=ics,
            context_classification=Node.INDIVIDUAL,
            task_classification=Node.ASSESS_PEERS,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1
        node = week.nodes.create(
            title="Review and change work based on feedback",
            description="Individually or in small groups, students review feedback and decide how to make changes.",
            column=ics,
            context_classification=Node.INDIVIDUAL,
            task_classification=Node.WRITE,
        )
        nodeweek = NodeWeek.objects.get(node=node)
        nodeweek.rank = rank
        nodeweek.save()
        rank = rank + 1

        for nodeweek in week.nodeweek_set.all():
            print(nodeweek.rank)



import json
import time

from django.contrib.auth.models import Group
from django.test import TestCase
from django.test.client import RequestFactory
from django.urls import reverse
from rest_framework.renderers import JSONRenderer

from course_flow import settings
from course_flow.models import (
    Activity,
    Course,
    LiveAssignment,
    LiveProject,
    LiveProjectUser,
    ObjectPermission,
    Project,
    User,
    UserAssignment,
    WorkflowProject,
)
from course_flow.serializers import LiveProjectSerializer

from .utils import get_author, login, login_student


class ModelViewTest(TestCase):
    def setUp(self):
        self.factory = RequestFactory()

    def test_home_view_student(self):
        response = self.client.get(reverse("course_flow:home"))
        self.assertEqual(response.status_code, 302)
        login_student(self)
        response = self.client.get(reverse("course_flow:home"))
        self.assertEqual(response.status_code, 200)

    def test_myclassrooms_view(self):
        response = self.client.get(reverse("course_flow:my-live-projects"))
        self.assertEqual(response.status_code, 302)
        login(self)
        response = self.client.get(reverse("course_flow:my-live-projects"))
        self.assertEqual(response.status_code, 200)

    def test_myclassrooms_view_student(self):
        Group.objects.create(name=settings.TEACHER_GROUP)
        response = self.client.get(reverse("course_flow:my-live-projects"))
        self.assertEqual(response.status_code, 302)
        login_student(self)
        response = self.client.get(reverse("course_flow:my-live-projects"))
        self.assertEqual(response.status_code, 200)

    def test_liveproject_update_view_owned(self):
        author = login(self)
        project = Project.objects.create(author=author)
        response = self.client.get(
            reverse("course_flow:live-project-update", args=[project.pk])
        )
        self.assertEqual(response.status_code, 403)
        LiveProject.objects.create(project=project)
        response = self.client.get(
            reverse("course_flow:live-project-update", args=[project.pk])
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            LiveProjectUser.objects.filter(
                user=author, role_type=LiveProjectUser.ROLE_TEACHER
            ).count(),
            1,
        )

    def test_liveproject_update_view_teacher(self):
        author = get_author()
        project = Project.objects.create(author=author)
        LiveProject.objects.create(project=project)
        response = self.client.get(
            reverse("course_flow:live-project-update", args=[project.pk])
        )
        self.assertEqual(response.status_code, 302)
        user = login(self)
        response = self.client.get(
            reverse("course_flow:live-project-update", args=[project.pk])
        )
        self.assertEqual(response.status_code, 403)
        LiveProjectUser.objects.create(
            liveproject=project.liveproject,
            user=user,
            role_type=LiveProjectUser.ROLE_TEACHER,
        )
        response = self.client.get(
            reverse("course_flow:live-project-update", args=[project.pk])
        )
        self.assertEqual(response.status_code, 200)

    def test_liveproject_update_view_student(self):
        author = get_author()
        project = Project.objects.create(author=author)
        LiveProject.objects.create(project=project)
        response = self.client.get(
            reverse("course_flow:live-project-update", args=[project.pk])
        )
        self.assertEqual(response.status_code, 302)
        user = login_student(self)
        response = self.client.get(
            reverse("course_flow:live-project-update", args=[project.pk])
        )
        self.assertEqual(response.status_code, 403)
        LiveProjectUser.objects.create(
            liveproject=project.liveproject,
            user=user,
            role_type=LiveProjectUser.ROLE_STUDENT,
        )
        response = self.client.get(
            reverse("course_flow:live-project-update", args=[project.pk])
        )
        self.assertEqual(response.status_code, 200)

    def test_liveproject_create_view(self):
        author = get_author()
        project = Project.objects.create(author=author)
        response = self.client.post(
            reverse("course_flow:make-project-live"),
            {
                "projectPk": project.id,
            },
        )
        self.assertEqual(response.status_code, 401)
        user = login(self)
        response = self.client.post(
            reverse("course_flow:make-project-live"),
            {
                "projectPk": project.id,
            },
        )
        self.assertEqual(response.status_code, 403)
        ObjectPermission.objects.create(
            user=user,
            content_object=project,
            permission_type=ObjectPermission.PERMISSION_EDIT,
        )
        response = self.client.post(
            reverse("course_flow:make-project-live"),
            {
                "projectPk": project.id,
            },
        )
        self.assertEqual(response.status_code, 403)
        project2 = Project.objects.create(author=user)
        response = self.client.post(
            reverse("course_flow:make-project-live"),
            {
                "projectPk": project2.id,
            },
        )
        self.assertEqual(response.status_code, 200)

    def test_liveproject_role_view(self):
        author = get_author()
        project = Project.objects.create(author=author)
        liveproject = LiveProject.objects.create(project=project)
        user = login(self)
        LiveProjectUser.objects.create(
            user=user,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_STUDENT,
        )
        # As a student, try to change your own role
        response = self.client.post(
            reverse("course_flow:set-liveproject-role"),
            {
                "liveprojectPk": project.id,
                "role_type": LiveProjectUser.ROLE_TEACHER,
                "permission_user": user.id,
            },
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            LiveProjectUser.objects.filter(
                user=user,
                liveproject=liveproject,
                role_type=LiveProjectUser.ROLE_TEACHER,
            ).count(),
            0,
        )
        LiveProjectUser.objects.create(
            user=user,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_TEACHER,
        )
        # As a teacher, try to set your own role
        response = self.client.post(
            reverse("course_flow:set-liveproject-role"),
            {
                "liveprojectPk": project.id,
                "role_type": LiveProjectUser.ROLE_TEACHER,
                "permission_user": user.id,
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            LiveProjectUser.objects.filter(
                user=user,
                liveproject=liveproject,
                role_type=LiveProjectUser.ROLE_TEACHER,
            ).count(),
            1,
        )
        user2 = User.objects.create()
        # As a teacher, try to change another user who has a student account's role to teacher
        response = self.client.post(
            reverse("course_flow:set-liveproject-role"),
            {
                "liveprojectPk": project.id,
                "role_type": LiveProjectUser.ROLE_TEACHER,
                "permission_user": user2.id,
            },
        )
        self.assertEqual(
            LiveProjectUser.objects.filter(
                user=user2,
                liveproject=liveproject,
                role_type=LiveProjectUser.ROLE_TEACHER,
            ).count(),
            0,
        )
        self.assertEqual(json.loads(response.content).get("action"), "error")
        # Try to change the author's permission
        response = self.client.post(
            reverse("course_flow:set-liveproject-role"),
            {
                "liveprojectPk": project.id,
                "role_type": str(LiveProjectUser.ROLE_TEACHER),
                "permission_user": author.id,
            },
        )
        self.assertEqual(
            LiveProjectUser.objects.filter(
                user=author,
                liveproject=liveproject,
                role_type=LiveProjectUser.ROLE_STUDENT,
            ).count(),
            0,
        )
        self.assertEqual(json.loads(response.content).get("action"), "error")

    def test_get_live_project_data(self):
        author = get_author()
        project = Project.objects.create(author=author)
        liveproject = LiveProject.objects.create(project=project)
        response = self.client.post(
            reverse("course_flow:get-live-project-data"),
            {
                "liveprojectPk": liveproject.pk,
                "data_type": JSONRenderer().render("overview").decode("utf-8"),
            },
        )
        self.assertEqual(response.status_code, 401)
        user = login(self)
        response = self.client.post(
            reverse("course_flow:get-live-project-data"),
            {
                "liveprojectPk": liveproject.pk,
                "data_type": JSONRenderer().render("overview").decode("utf-8"),
            },
        )
        self.assertEqual(response.status_code, 403)
        LiveProjectUser.objects.create(
            user=user,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_STUDENT,
        )
        response = self.client.post(
            reverse("course_flow:get-live-project-data"),
            {
                "liveprojectPk": liveproject.pk,
                "data_type": JSONRenderer().render("overview").decode("utf-8"),
            },
        )
        self.assertEqual(response.status_code, 403)
        LiveProjectUser.objects.create(
            user=user,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_TEACHER,
        )
        response = self.client.post(
            reverse("course_flow:get-live-project-data"),
            {
                "liveprojectPk": liveproject.pk,
                "data_type": JSONRenderer().render("overview").decode("utf-8"),
            },
        )
        self.assertEqual(response.status_code, 200)

    def test_get_users_for_liveproject(self):
        author = get_author()
        project = Project.objects.create(author=author)
        liveproject = LiveProject.objects.create(project=project)
        response = self.client.post(
            reverse("course_flow:get-users-for-liveproject"),
            {
                "liveprojectPk": liveproject.pk,
            },
        )
        self.assertEqual(response.status_code, 401)
        user = login(self)
        response = self.client.post(
            reverse("course_flow:get-users-for-liveproject"),
            {
                "liveprojectPk": liveproject.pk,
            },
        )
        self.assertEqual(response.status_code, 403)
        LiveProjectUser.objects.create(
            user=user,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_STUDENT,
        )
        response = self.client.post(
            reverse("course_flow:get-users-for-liveproject"),
            {
                "liveprojectPk": liveproject.pk,
            },
        )
        self.assertEqual(response.status_code, 403)
        LiveProjectUser.objects.create(
            user=user,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_TEACHER,
        )
        response = self.client.post(
            reverse("course_flow:get-users-for-liveproject"),
            {
                "liveprojectPk": liveproject.pk,
            },
        )
        self.assertEqual(response.status_code, 200)

    def test_register_as_student(self):
        author = get_author()
        project = Project.objects.create(author=author)
        liveproject = LiveProject.objects.create(project=project)
        response = self.client.get(
            reverse(
                "course_flow:register-as-student",
                args=[project.registration_hash()],
            ),
        )
        self.assertEqual(response.status_code, 401)
        user = login_student(self)
        response = self.client.post(
            reverse(
                "course_flow:register-as-student",
                args=[project.registration_hash()],
            ),
        )
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            LiveProjectUser.objects.filter(
                user=user,
                liveproject=liveproject,
                role_type=LiveProjectUser.ROLE_STUDENT,
            ).count(),
            1,
        )

    def test_register_as_owner(self):
        author = login(self)
        project = Project.objects.create(author=author)
        liveproject = LiveProject.objects.create(project=project)
        response = self.client.post(
            reverse(
                "course_flow:register-as-student",
                args=[project.registration_hash()],
            ),
        )
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            LiveProjectUser.objects.filter(
                user=author,
                liveproject=liveproject,
                role_type=LiveProjectUser.ROLE_STUDENT,
            ).count(),
            0,
        )
        self.assertEqual(
            LiveProjectUser.objects.filter(
                user=author,
                liveproject=liveproject,
                role_type=LiveProjectUser.ROLE_TEACHER,
            ).count(),
            1,
        )

    def test_project_exposes_hash(self):
        author = get_author()
        project = Project.objects.create(author=author)
        liveproject = LiveProject.objects.create(project=project)
        user = User.objects.create()
        serialized = LiveProjectSerializer(
            liveproject, context={"user": author}
        ).data
        self.assertEqual(
            serialized.get("registration_hash"), project.registration_hash()
        )
        serialized = LiveProjectSerializer(
            liveproject, context={"user": user}
        ).data
        self.assertNotEqual(
            serialized.get("registration_hash"), project.registration_hash()
        )

    def test_liveproject_workflow_visibility_view(self):
        author = get_author()
        project = Project.objects.create(author=author)
        liveproject = LiveProject.objects.create(project=project)
        workflow = Activity.objects.create(author=author)
        WorkflowProject.objects.create(project=project, workflow=workflow)
        user = login(self)
        response = self.client.post(
            reverse("course_flow:set-workflow-visibility"),
            {
                "liveprojectPk": project.id,
                "workflowPk": workflow.pk,
                "visible": JSONRenderer().render(True).decode("utf-8"),
            },
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(liveproject.visible_workflows.count(), 0)

        LiveProjectUser.objects.create(
            user=user,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_STUDENT,
        )
        response = self.client.post(
            reverse("course_flow:set-workflow-visibility"),
            {
                "liveprojectPk": project.id,
                "workflowPk": workflow.pk,
                "visible": JSONRenderer().render(True).decode("utf-8"),
            },
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(liveproject.visible_workflows.count(), 0)

        ObjectPermission.objects.create(
            user=user,
            content_object=workflow,
            permission_type=ObjectPermission.PERMISSION_VIEW,
        )
        response = self.client.post(
            reverse("course_flow:set-workflow-visibility"),
            {
                "liveprojectPk": project.id,
                "workflowPk": workflow.pk,
                "visible": JSONRenderer().render(True).decode("utf-8"),
            },
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(liveproject.visible_workflows.count(), 0)

        ObjectPermission.objects.create(
            user=user,
            content_object=workflow,
            permission_type=ObjectPermission.PERMISSION_NONE,
        )
        LiveProjectUser.objects.create(
            user=user,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_TEACHER,
        )
        response = self.client.post(
            reverse("course_flow:set-workflow-visibility"),
            {
                "liveprojectPk": project.id,
                "workflowPk": workflow.pk,
                "visible": JSONRenderer().render(True).decode("utf-8"),
            },
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(liveproject.visible_workflows.count(), 0)

        ObjectPermission.objects.create(
            user=user,
            content_object=workflow,
            permission_type=ObjectPermission.PERMISSION_VIEW,
        )
        response = self.client.post(
            reverse("course_flow:set-workflow-visibility"),
            {
                "liveprojectPk": project.id,
                "workflowPk": workflow.pk,
                "visible": JSONRenderer().render(True).decode("utf-8"),
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(liveproject.visible_workflows.count(), 1)

        # try to add a workflow from a different project, just for fun
        workflow2 = Activity.objects.create(author=user)
        WorkflowProject.objects.create(
            project=Project.objects.create(author=user), workflow=workflow2
        )
        response = self.client.post(
            reverse("course_flow:set-workflow-visibility"),
            {
                "liveprojectPk": project.id,
                "workflowPk": workflow2.pk,
                "visible": JSONRenderer().render(True).decode("utf-8"),
            },
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(liveproject.visible_workflows.count(), 1)

    def test_access_workflow_as_student_teacher(self):
        author = get_author()
        project = Project.objects.create(author=author)
        liveproject = LiveProject.objects.create(project=project)
        workflow = Activity.objects.create(author=author)
        WorkflowProject.objects.create(project=project, workflow=workflow)
        user = login(self)

        LiveProjectUser.objects.create(
            user=user,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_STUDENT,
        )
        response = self.client.get(
            reverse("course_flow:workflow-update", args=[workflow.pk])
        )
        self.assertEqual(response.status_code, 403)
        response = self.client.post(
            reverse("course_flow:get-workflow-data"),
            {
                "workflowPk": workflow.pk,
            },
        )
        self.assertEqual(response.status_code, 403)

        LiveProjectUser.objects.create(
            user=user,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_TEACHER,
        )
        response = self.client.get(
            reverse("course_flow:workflow-update", args=[workflow.pk])
        )
        self.assertEqual(response.status_code, 403)
        response = self.client.post(
            reverse("course_flow:get-workflow-data"),
            {
                "workflowPk": workflow.pk,
            },
        )
        self.assertEqual(response.status_code, 403)

        liveproject.visible_workflows.add(workflow)

        LiveProjectUser.objects.create(
            user=user,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_STUDENT,
        )
        response = self.client.get(
            reverse("course_flow:workflow-update", args=[workflow.pk])
        )
        self.assertEqual(response.status_code, 200)
        response = self.client.post(
            reverse("course_flow:get-workflow-data"),
            {
                "workflowPk": workflow.pk,
            },
        )
        self.assertEqual(response.status_code, 200)

        LiveProjectUser.objects.create(
            user=user,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_TEACHER,
        )
        response = self.client.get(
            reverse("course_flow:workflow-update", args=[workflow.pk])
        )
        self.assertEqual(response.status_code, 200)
        response = self.client.post(
            reverse("course_flow:get-workflow-data"),
            {
                "workflowPk": workflow.pk,
            },
        )
        self.assertEqual(response.status_code, 200)

    def test_assignment_create_view(self):
        author = get_author()
        project = Project.objects.create(author=author)
        workflow = Course.objects.create(author=author)
        WorkflowProject.objects.create(workflow=workflow, project=project)
        column = workflow.columns.create(author=author)  # noqa F841
        week = workflow.weeks.create(author=author)
        node = week.nodes.create(author=author)

        liveproject = LiveProject.objects.create(project=project)
        response = self.client.post(
            reverse("course_flow:create-live-assignment"),
            {"liveprojectPk": liveproject.pk, "nodePk": node.pk},
        )
        self.assertEqual(response.status_code, 401)
        user = login(self)
        response = self.client.post(
            reverse("course_flow:create-live-assignment"),
            {"liveprojectPk": liveproject.pk, "nodePk": node.pk},
        )
        self.assertEqual(response.status_code, 403)
        LiveProjectUser.objects.create(
            user=user,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_STUDENT,
        )
        response = self.client.post(
            reverse("course_flow:create-live-assignment"),
            {"liveprojectPk": liveproject.pk, "nodePk": node.pk},
        )
        self.assertEqual(response.status_code, 403)
        LiveProjectUser.objects.create(
            user=user,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_TEACHER,
        )
        response = self.client.post(
            reverse("course_flow:create-live-assignment"),
            {"liveprojectPk": liveproject.pk, "nodePk": node.pk},
        )
        self.assertEqual(response.status_code, 403)
        liveproject.visible_workflows.add(workflow)
        response = self.client.post(
            reverse("course_flow:create-live-assignment"),
            {"liveprojectPk": liveproject.pk, "nodePk": node.pk},
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(LiveAssignment.objects.all().count(), 1)

    def test_assignment_add_users_view(self):
        author = get_author()
        project = Project.objects.create(author=author)
        workflow = Course.objects.create(author=author)
        WorkflowProject.objects.create(workflow=workflow, project=project)
        column = workflow.columns.create(author=author)  # noqa F841
        week = workflow.weeks.create(author=author)
        node = week.nodes.create(author=author)

        liveproject = LiveProject.objects.create(project=project)
        assignment = LiveAssignment.objects.create(
            liveproject=liveproject, task=node
        )
        LiveProjectUser.objects.create(
            user=author,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_TEACHER,
        )
        user_list = [author.pk]
        response = self.client.post(
            reverse("course_flow:add-users-to-assignment"),
            {
                "liveassignmentPk": assignment.pk,
                "user_list": JSONRenderer().render(user_list).decode("utf-8"),
                "add": JSONRenderer().render(True).decode("utf-8"),
            },
        )
        self.assertEqual(response.status_code, 401)
        user = login(self)
        response = self.client.post(
            reverse("course_flow:add-users-to-assignment"),
            {
                "liveassignmentPk": assignment.pk,
                "user_list": JSONRenderer().render(user_list).decode("utf-8"),
                "add": JSONRenderer().render(True).decode("utf-8"),
            },
        )
        self.assertEqual(response.status_code, 403)
        LiveProjectUser.objects.create(
            user=user,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_STUDENT,
        )
        response = self.client.post(
            reverse("course_flow:add-users-to-assignment"),
            {
                "liveassignmentPk": assignment.pk,
                "user_list": JSONRenderer().render(user_list).decode("utf-8"),
                "add": JSONRenderer().render(True).decode("utf-8"),
            },
        )
        self.assertEqual(response.status_code, 403)
        LiveProjectUser.objects.create(
            user=user,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_TEACHER,
        )
        response = self.client.post(
            reverse("course_flow:add-users-to-assignment"),
            {
                "liveassignmentPk": assignment.pk,
                "user_list": JSONRenderer().render(user_list).decode("utf-8"),
                "add": JSONRenderer().render(True).decode("utf-8"),
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            UserAssignment.objects.filter(
                user=user_list[0], assignment=assignment
            ).count(),
            1,
        )
        response = self.client.post(
            reverse("course_flow:add-users-to-assignment"),
            {
                "liveassignmentPk": assignment.pk,
                "user_list": JSONRenderer().render(user_list).decode("utf-8"),
                "add": JSONRenderer().render(False).decode("utf-8"),
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            UserAssignment.objects.filter(
                user=user_list[0], assignment=assignment
            ).count(),
            0,
        )

    def test_set_assignment_completion_view(self):
        author = get_author()
        project = Project.objects.create(author=author)
        workflow = Course.objects.create(author=author)
        WorkflowProject.objects.create(workflow=workflow, project=project)
        column = workflow.columns.create(author=author)  # noqa F841
        week = workflow.weeks.create(author=author)
        node = week.nodes.create(author=author)

        liveproject = LiveProject.objects.create(
            project=project, default_assign_to_all=False
        )
        assignment = LiveAssignment.objects.create(
            liveproject=liveproject, task=node
        )
        LiveProjectUser.objects.create(
            user=author,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_TEACHER,
        )
        userassignment = UserAssignment.objects.create(
            user=author,
            assignment=assignment,
        )
        response = self.client.post(
            reverse("course_flow:set-assignment-completion"),
            {
                "userassignmentPk": userassignment.pk,
                "completed": JSONRenderer().render(True).decode("utf-8"),
            },
        )
        self.assertEqual(response.status_code, 401)
        user = login(self)
        response = self.client.post(
            reverse("course_flow:set-assignment-completion"),
            {
                "userassignmentPk": userassignment.pk,
                "completed": JSONRenderer().render(True).decode("utf-8"),
            },
        )
        self.assertEqual(response.status_code, 403)
        lpu = LiveProjectUser.objects.create(
            user=user,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_STUDENT,
        )
        response = self.client.post(
            reverse("course_flow:set-assignment-completion"),
            {
                "userassignmentPk": userassignment.pk,
                "completed": JSONRenderer().render(True).decode("utf-8"),
            },
        )
        self.assertEqual(
            UserAssignment.objects.get(user=author).completed, False
        )
        self.assertEqual(response.status_code, 403)
        userassignment2 = UserAssignment.objects.create(
            user=user,
            assignment=assignment,
        )
        response = self.client.post(
            reverse("course_flow:set-assignment-completion"),
            {
                "userassignmentPk": userassignment2.pk,
                "completed": JSONRenderer().render(True).decode("utf-8"),
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(UserAssignment.objects.get(user=user).completed, True)
        assignment.self_reporting = False
        assignment.save()
        response = self.client.post(
            reverse("course_flow:set-assignment-completion"),
            {
                "userassignmentPk": userassignment2.pk,
                "completed": JSONRenderer().render(False).decode("utf-8"),
            },
        )
        self.assertEqual(response.status_code, 403)
        lpu.role_type = LiveProjectUser.ROLE_TEACHER
        lpu.save()
        response = self.client.post(
            reverse("course_flow:set-assignment-completion"),
            {
                "userassignmentPk": userassignment2.pk,
                "completed": JSONRenderer().render(False).decode("utf-8"),
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            UserAssignment.objects.get(user=user).completed, False
        )
        response = self.client.post(
            reverse("course_flow:set-assignment-completion"),
            {
                "userassignmentPk": userassignment.pk,
                "completed": JSONRenderer().render(True).decode("utf-8"),
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            UserAssignment.objects.get(user=author).completed, True
        )

    def test_get_assignments_for_node(self):
        author = get_author()
        project = Project.objects.create(author=author)
        workflow = Course.objects.create(author=author)
        WorkflowProject.objects.create(workflow=workflow, project=project)
        column = workflow.columns.first()
        week = workflow.weeks.create(author=author)
        node = week.nodes.create(author=author)
        node.column = column
        node.save()

        # Create a liveproject with an assignment
        liveproject = LiveProject.objects.create(project=project)
        assignment = LiveAssignment.objects.create(
            liveproject=liveproject, task=node
        )
        # Add the author as a liveprojectuser teacher
        LiveProjectUser.objects.create(
            user=author,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_TEACHER,
        )
        # Assign the assignment to the author
        userassignment = UserAssignment.objects.create(
            user=author,
            assignment=assignment,
        )
        response = self.client.post(
            reverse("course_flow:get-assignments-for-node"),
            {
                "nodePk": node.pk,
            },
        )
        self.assertEqual(response.status_code, 401)
        user = login(self)
        response = self.client.post(
            reverse("course_flow:get-assignments-for-node"),
            {
                "nodePk": node.pk,
            },
        )
        self.assertEqual(response.status_code, 403)
        # Add the user to the project
        lpu = LiveProjectUser.objects.create(
            user=user,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_STUDENT,
        )
        response = self.client.post(
            reverse("course_flow:get-assignments-for-node"),
            {
                "nodePk": node.pk,
            },
        )
        self.assertEqual(response.status_code, 403)

        liveproject.visible_workflows.add(workflow)
        # Add the user to the assignment - not needed always because happens by default
        if UserAssignment.objects.filter(user=user).count() == 0:
            UserAssignment.objects.create(
                user=user,
                assignment=assignment,
            )
        self.assertEqual(len(UserAssignment.objects.filter(user=user)),1)
        response = self.client.post(
            reverse("course_flow:get-assignments-for-node"),
            {
                "nodePk": node.pk,
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(json.loads(response.content)["action"], "posted")
        self.assertEqual(
            len(
                json.loads(response.content)["data_package"]["my_assignments"]
            ),
            1,
        )
        self.assertEqual(
            json.loads(response.content)["data_package"]["all_assignments"],
            None,
        )

        UserAssignment.objects.filter(user=user).delete()
        response = self.client.post(
            reverse("course_flow:get-assignments-for-node"),
            {
                "nodePk": node.pk,
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(json.loads(response.content)["action"], "posted")
        self.assertEqual(
            len(
                json.loads(response.content)["data_package"]["my_assignments"]
            ),
            0,
        )
        self.assertEqual(
            json.loads(response.content)["data_package"]["all_assignments"],
            None,
        )

        userassignment = UserAssignment.objects.create(
            user=user,
            assignment=assignment,
        )
        lpu = LiveProjectUser.objects.create(
            user=user,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_TEACHER,
        )
        response = self.client.post(
            reverse("course_flow:get-assignments-for-node"),
            {
                "nodePk": node.pk,
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(json.loads(response.content)["action"], "posted")
        self.assertEqual(
            len(
                json.loads(response.content)["data_package"]["my_assignments"]
            ),
            1,
        )
        self.assertEqual(
            len(
                json.loads(response.content)["data_package"]["all_assignments"]
            ),
            1,
        )

        UserAssignment.objects.filter(user=user).delete()
        response = self.client.post(
            reverse("course_flow:get-assignments-for-node"),
            {
                "nodePk": node.pk,
            },
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(json.loads(response.content)["action"], "posted")
        self.assertEqual(
            len(
                json.loads(response.content)["data_package"]["my_assignments"]
            ),
            0,
        )
        self.assertEqual(
            len(
                json.loads(response.content)["data_package"]["all_assignments"]
            ),
            1,
        )

    def test_assignment_default_values(self):
        author = get_author()
        student = get_author("3")
        teacher = get_author("4")
        project = Project.objects.create(author=author)
        workflow = Course.objects.create(author=author)
        WorkflowProject.objects.create(workflow=workflow, project=project)
        column = workflow.columns.first()
        week = workflow.weeks.first()
        node = week.nodes.create(author=author)
        node.column = column
        node.save()
        liveproject = LiveProject.objects.create(project=project)

        LiveProjectUser.objects.create(
            user=student,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_STUDENT,
        )
        LiveProjectUser.objects.create(
            user=teacher,
            liveproject=liveproject,
            role_type=LiveProjectUser.ROLE_TEACHER,
        )

        assignment = LiveAssignment.objects.create(
            liveproject=liveproject, task=node
        )

        self.assertEqual(
            assignment.self_reporting, liveproject.default_self_reporting
        )
        self.assertEqual(
            assignment.single_completion, liveproject.default_single_completion
        )
        self.assertEqual(
            UserAssignment.objects.filter(
                user=student, assignment=assignment
            ).count(),
            1,
        )
        self.assertEqual(
            UserAssignment.objects.filter(
                user=teacher, assignment=assignment
            ).count(),
            0,
        )

        liveproject.default_single_completion = True
        liveproject.default_self_reporting = False
        liveproject.default_assign_to_all = False

        liveproject.save()

        assignment = LiveAssignment.objects.create(
            liveproject=liveproject, task=node
        )

        self.assertEqual(
            assignment.self_reporting, liveproject.default_self_reporting
        )
        self.assertEqual(
            assignment.single_completion, liveproject.default_single_completion
        )
        self.assertEqual(
            UserAssignment.objects.filter(
                user=student, assignment=assignment
            ).count(),
            0,
        )
        self.assertEqual(
            UserAssignment.objects.filter(
                user=teacher, assignment=assignment
            ).count(),
            0,
        )



# Generated by Django 3.2.15 on 2023-02-05 16:12

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


def forwards_func(apps, schema_editor):
    Activity = apps.get_model("course_flow", "Activity")
    Course = apps.get_model("course_flow", "Course")
    Program = apps.get_model("course_flow", "Program")
    for workflow in (
        list(Activity.objects.all())
        + list(Course.objects.all())
        + list(Program.objects.all())
    ):
        workflow.author_temp = workflow.author
        workflow.save()


def reverse_func(apps, schema_editor):
    Activity = apps.get_model("course_flow", "Activity")
    Course = apps.get_model("course_flow", "Course")
    Program = apps.get_model("course_flow", "Program")
    for workflow in (
        list(Activity.objects.all())
        + list(Course.objects.all())
        + list(Program.objects.all())
    ):
        workflow.author = workflow.author_temp
        workflow.save()


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("course_flow", "0101_auto_20230118_2202"),
    ]

    operations = [
        migrations.AddField(
            model_name="workflow",
            name="author_temp",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="authored_workflows",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.RunPython(forwards_func, reverse_func),
        migrations.AlterField(
            model_name="courseflowuser",
            name="notifications",
            field=models.BooleanField(
                default=False,
                help_text="Check this box if you would like to receive emails from us about updates to CourseFlow.",
            ),
        ),
    ]



import base64
import uuid

from django.contrib.auth import get_user_model
from django.contrib.contenttypes.fields import (
    GenericForeignKey,
    GenericRelation,
)
from django.contrib.contenttypes.models import ContentType
from django.core.cache import cache
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q
from django.db.models.signals import post_save, pre_delete, pre_save
from django.dispatch import receiver
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from model_utils.managers import InheritanceManager

from course_flow.utils import (
    get_all_outcomes_for_outcome,
    get_descendant_outcomes,
)

User = get_user_model()

title_max_length = 200


class Project(models.Model):
    deleted = models.BooleanField(default=False)
    deleted_on = models.DateTimeField(default=timezone.now)
    title = models.CharField(
        max_length=title_max_length, null=True, blank=True
    )
    description = models.TextField(null=True, blank=True)
    author = models.ForeignKey(
        User,
        related_name="authored_projects",
        on_delete=models.SET_NULL,
        null=True,
    )
    created_on = models.DateTimeField(default=timezone.now)
    last_modified = models.DateTimeField(auto_now=True)
    published = models.BooleanField(default=False)

    from_saltise = models.BooleanField(default=False)

    is_strategy = models.BooleanField(default=False)

    workflows = models.ManyToManyField(
        "Workflow", through="WorkflowProject", blank=True
    )

    is_original = models.BooleanField(default=False)
    parent_project = models.ForeignKey(
        "Project", on_delete=models.SET_NULL, null=True
    )

    disciplines = models.ManyToManyField("Discipline", blank=True)

    favourited_by = GenericRelation("Favourite", related_query_name="project")
    notified_by = GenericRelation("Notification", related_query_name="project")
    user_permissions = GenericRelation(
        "ObjectPermission", related_query_name="project"
    )

    object_sets = models.ManyToManyField("ObjectSet", blank=True)

    hash = models.UUIDField(default=uuid.uuid4, editable=False, unique=True)

    @staticmethod
    def get_from_hash(hash_):
        try:
            hash_ = base64.urlsafe_b64decode(hash_.encode()).decode()
        except UnicodeDecodeError:
            hash_ = None
        if hash_:
            try:
                project = Project.objects.get(hash=hash_)
            except (Project.DoesNotExist, ValidationError):
                project = None
        else:
            project = None

        return project

    def registration_hash(self):
        return base64.urlsafe_b64encode(str(self.hash).encode()).decode()

    @property
    def type(self):
        return "project"

    def get_project(self):
        return self

    def get_live_project(self):
        try:
            liveproject = self.liveproject
        except AttributeError:
            liveproject = None
        return liveproject

    def get_permission_objects(self):
        return [self]

    def __str__(self):
        if self.title is not None and self.title != "":
            return self.title
        else:
            return "Project"

    class Meta:
        verbose_name = _("Project")
        verbose_name_plural = _("Projects")


class ObjectSet(models.Model):
    term = models.CharField(max_length=title_max_length)
    title = models.CharField(max_length=title_max_length)
    translation_plural = models.CharField(
        max_length=title_max_length, null=True
    )

    def get_permission_objects(self):
        return [Project.objects.filter(object_sets=self).first()]


class WorkflowProject(models.Model):
    project = models.ForeignKey(Project, on_delete=models.CASCADE)
    workflow = models.ForeignKey("Workflow", on_delete=models.CASCADE)
    added_on = models.DateTimeField(default=timezone.now)
    rank = models.PositiveIntegerField(default=0)

    def get_permission_objects(self):
        return [self.project, self.get_workflow().get_permission_objects()[0]]

    class Meta:
        verbose_name = _("Workflow-Project Link")
        verbose_name_plural = _("Workflow-Project Links")


# class OutcomeProject(models.Model):
#    project = models.ForeignKey(Project, on_delete=models.CASCADE)
#    outcome = models.ForeignKey("Outcome", on_delete=models.CASCADE)
#    added_on = models.DateTimeField(default=timezone.now)
#    rank = models.PositiveIntegerField(default=0)
#
#    def get_permission_objects(self):
#        return [self.project, self.outcome]
#
#    class Meta:
#        verbose_name = "Outcome-Project Link"
#        verbose_name_plural = "Outcome-Project Links"


class OutcomeWorkflow(models.Model):
    workflow = models.ForeignKey("Workflow", on_delete=models.CASCADE)
    outcome = models.ForeignKey("Outcome", on_delete=models.CASCADE)
    added_on = models.DateTimeField(default=timezone.now)
    rank = models.PositiveIntegerField(default=0)

    def get_display_rank(self):
        if self.outcome.deleted:
            return -1
        return list(
            OutcomeWorkflow.objects.filter(
                workflow=self.workflow, outcome__deleted=False
            ).order_by("rank")
        ).index(self)

    def get_permission_objects(self):
        return [self.project, self.outcome]

    class Meta:
        verbose_name = _("Outcome-Workflow Link")
        verbose_name_plural = _("Outcome-Workflow Links")


class Column(models.Model):
    deleted = models.BooleanField(default=False)
    deleted_on = models.DateTimeField(default=timezone.now)
    title = models.CharField(
        max_length=title_max_length, null=True, blank=True
    )
    author = models.ForeignKey(User, on_delete=models.SET_NULL, null=True)
    icon = models.CharField(max_length=50, null=True, blank=True)
    created_on = models.DateTimeField(default=timezone.now)
    last_modified = models.DateTimeField(auto_now=True)
    visible = models.BooleanField(default=True)
    colour = models.PositiveIntegerField(null=True)
    CUSTOM_ACTIVITY = 0
    OUT_OF_CLASS_INSTRUCTOR = 1
    OUT_OF_CLASS_STUDENT = 2
    IN_CLASS_INSTRUCTOR = 3
    IN_CLASS_STUDENT = 4
    CUSTOM_COURSE = 10
    PREPARATION = 11
    LESSON = 12
    ARTIFACT = 13
    ASSESSMENT = 14
    CUSTOM_PROGRAM = 20

    COLUMN_TYPES = (
        (CUSTOM_ACTIVITY, _("Custom Activity Column")),
        (OUT_OF_CLASS_INSTRUCTOR, _("Out of Class (Instructor)")),
        (OUT_OF_CLASS_STUDENT, _("Out of Class (Students)")),
        (IN_CLASS_INSTRUCTOR, _("In Class (Instructor)")),
        (IN_CLASS_STUDENT, _("In Class (Students)")),
        (CUSTOM_COURSE, _("Custom Course Column")),
        (PREPARATION, _("Preparation")),
        (LESSON, _("Lesson")),
        (ARTIFACT, _("Artifact")),
        (ASSESSMENT, _("Assessment")),
        (CUSTOM_PROGRAM, _("Custom Program Category")),
    )
    column_type = models.PositiveIntegerField(default=0, choices=COLUMN_TYPES)

    is_original = models.BooleanField(default=False)
    parent_column = models.ForeignKey(
        "Column", on_delete=models.SET_NULL, null=True
    )

    comments = models.ManyToManyField(
        "Comment", blank=True, related_name="column"
    )

    hash = models.UUIDField(default=uuid.uuid4, editable=False, unique=True)

    def get_permission_objects(self):
        return self.get_workflow().get_permission_objects()

    def get_workflow(self):
        return self.workflow_set.first()

    def get_display_title(self):
        if self.title is not None and self.title != "":
            return self.title
        else:
            return self.get_column_type_display()

    def __str__(self):
        return self.get_column_type_display()

    class Meta:
        verbose_name = _("Column")
        verbose_name_plural = _("Columns")


class NodeLink(models.Model):
    deleted = models.BooleanField(default=False)
    deleted_on = models.DateTimeField(default=timezone.now)
    title = models.CharField(max_length=100, null=True, blank=True)
    author = models.ForeignKey(User, on_delete=models.SET_NULL, null=True)
    source_node = models.ForeignKey(
        "Node", on_delete=models.CASCADE, related_name="outgoing_links"
    )
    target_node = models.ForeignKey(
        "Node", on_delete=models.CASCADE, related_name="incoming_links"
    )
    NORTH = 0
    EAST = 1
    SOUTH = 2
    WEST = 3
    SOURCE_PORTS = ((EAST, "e"), (SOUTH, "s"), (WEST, "w"))
    TARGET_PORTS = ((NORTH, "n"), (EAST, "e"), (WEST, "w"))
    source_port = models.PositiveIntegerField(choices=SOURCE_PORTS, default=2)
    target_port = models.PositiveIntegerField(choices=TARGET_PORTS, default=0)

    dashed = models.BooleanField(default=False)
    text_position = models.PositiveSmallIntegerField(default=20)
    created_on = models.DateTimeField(default=timezone.now)
    last_modified = models.DateTimeField(auto_now=True)

    is_original = models.BooleanField(default=True)

    def get_permission_objects(self):
        return self.get_workflow().get_permission_objects()

    def get_workflow(self):
        return self.source_node.get_workflow()

    hash = models.UUIDField(default=uuid.uuid4, editable=False, unique=True)

    class Meta:
        verbose_name = _("Node Link")
        verbose_name_plural = _("Node Links")


class Outcome(models.Model):
    deleted = models.BooleanField(default=False)
    deleted_on = models.DateTimeField(default=timezone.now)
    title = models.TextField(null=True, blank=True)
    code = models.CharField(max_length=title_max_length, null=True, blank=True)
    description = models.TextField(null=True, blank=True)
    author = models.ForeignKey(User, on_delete=models.SET_NULL, null=True)
    created_on = models.DateTimeField(default=timezone.now)
    last_modified = models.DateTimeField(auto_now=True)
    parent_outcome = models.ForeignKey(
        "Outcome", on_delete=models.SET_NULL, null=True
    )
    is_original = models.BooleanField(default=True)

    is_dropped = models.BooleanField(default=True)
    depth = models.PositiveIntegerField(default=0)

    sets = models.ManyToManyField("ObjectSet", blank=True)

    children = models.ManyToManyField(
        "Outcome",
        through="OutcomeOutcome",
        blank=True,
        related_name="parent_outcomes",
    )

    horizontal_outcomes = models.ManyToManyField(
        "Outcome",
        through="OutcomeHorizontalLink",
        blank=True,
        related_name="reverse_horizontal_outcomes",
    )

    comments = models.ManyToManyField(
        "Comment", blank=True, related_name="outcome"
    )

    hash = models.UUIDField(default=uuid.uuid4, editable=False, unique=True)

    def get_top_outcome(self):
        if self.parent_outcome_links.all().count() > 0:
            return self.parent_outcome_links.first().parent.get_top_outcome()
        else:
            return self

    def get_workflow(self):
        return self.get_top_outcome().workflow_set.first()

    #
    #    def get_project(self):
    #        return self.project_set.first()

    def get_permission_objects(self):
        return self.get_workflow().get_permission_objects()

    def __str__(self):
        return self.title

    def get_all_outcome_ids(self, ids):
        ids.append(self.id)
        for outcome in self.children.all():
            outcome.get_all_outcome_ids(ids)

    class Meta:
        verbose_name = _("Outcome")
        verbose_name_plural = _("Outcomes")


class OutcomeOutcome(models.Model):
    parent = models.ForeignKey(
        Outcome, on_delete=models.CASCADE, related_name="child_outcome_links"
    )
    child = models.ForeignKey(
        Outcome, on_delete=models.CASCADE, related_name="parent_outcome_links"
    )
    added_on = models.DateTimeField(default=timezone.now)
    rank = models.PositiveIntegerField(default=0)

    def get_display_rank(self):
        if self.child.deleted:
            return -1
        return list(
            OutcomeOutcome.objects.filter(
                parent=self.parent, child__deleted=False
            ).order_by("rank")
        ).index(self)

    def get_permission_objects(self):
        return self.get_top_outcome().get_permission_objects()

    def get_top_outcome(self):
        return self.parent.get_top_outcome()

    class Meta:
        verbose_name = _("Outcome-Outcome Link")
        verbose_name_plural = _("Outcome-Outcome Links")


class OutcomeHorizontalLink(models.Model):
    outcome = models.ForeignKey(
        Outcome,
        on_delete=models.CASCADE,
        related_name="outcome_horizontal_links",
    )
    parent_outcome = models.ForeignKey(
        Outcome,
        on_delete=models.CASCADE,
        related_name="reverse_outcome_horizontal_links",
    )
    added_on = models.DateTimeField(default=timezone.now)
    rank = models.PositiveIntegerField(default=0)
    degree = models.PositiveIntegerField(default=1)

    def get_permission_objects(self):
        return self.get_top_outcome().get_permission_objects()

    def get_top_outcome(self):
        return self.outcome.get_top_outcome()

    # Check to see if the parent has all its children the same, and add it if necessary
    def check_parent_outcomes(self):
        if self.parent_outcome.parent_outcomes.count() > 0:
            parent_outcome = self.parent_outcome.parent_outcomes.first()
            if (
                OutcomeHorizontalLink.objects.filter(
                    parent_outcome__in=parent_outcome.children.exclude(
                        deleted=True
                    ).values_list("id", flat=True),
                    degree=self.degree,
                    outcome=self.outcome,
                ).count()
                == parent_outcome.children.exclude(deleted=True).count()
            ):
                new_outcomehorizontallink = (
                    OutcomeHorizontalLink.objects.create(
                        outcome=self.outcome,
                        degree=self.degree,
                        parent_outcome=parent_outcome,
                    )
                )
                return [
                    new_outcomehorizontallink
                ] + new_outcomehorizontallink.check_parent_outcomes()
            elif (
                OutcomeHorizontalLink.objects.filter(
                    parent_outcome=parent_outcome, outcome=self.outcome
                ).count()
                > 0
            ):
                new_outcomehorizontallink = (
                    OutcomeHorizontalLink.objects.create(
                        outcome=self.outcome,
                        degree=0,
                        parent_outcome=parent_outcome,
                    )
                )
                return [
                    new_outcomehorizontallink
                ] + new_outcomehorizontallink.check_parent_outcomes()

        return []

    # Check to see if the children already exist, and if not, add them
    def check_child_outcomes(self):
        new_children = []
        for child in self.parent_outcome.children.all():
            if (
                OutcomeHorizontalLink.objects.filter(
                    parent_outcome=child,
                    outcome=self.outcome,
                    degree=self.degree,
                ).count()
                == 0
            ):
                new_child = OutcomeHorizontalLink.objects.create(
                    parent_outcome=child,
                    outcome=self.outcome,
                    degree=self.degree,
                )
                new_children += [new_child] + new_child.check_child_outcomes()
        return new_children

    class Meta:
        verbose_name = _("Outcome-Outcome Link")
        verbose_name_plural = _("Outcome-Outcome Links")


class Node(models.Model):
    deleted = models.BooleanField(default=False)
    deleted_on = models.DateTimeField(default=timezone.now)
    title = models.CharField(
        max_length=title_max_length, null=True, blank=True
    )
    description = models.TextField(null=True, blank=True)
    author = models.ForeignKey(
        User,
        related_name="authored_nodes",
        on_delete=models.SET_NULL,
        null=True,
    )
    created_on = models.DateTimeField(default=timezone.now)
    last_modified = models.DateTimeField(auto_now=True)

    parent_node = models.ForeignKey(
        "Node", on_delete=models.SET_NULL, null=True
    )
    is_original = models.BooleanField(default=True)
    has_autolink = models.BooleanField(default=False)
    is_dropped = models.BooleanField(default=False)

    comments = models.ManyToManyField(
        "Comment", blank=True, related_name="node"
    )

    sets = models.ManyToManyField("ObjectSet", blank=True)

    NONE = 0
    INDIVIDUAL = 1
    GROUPS = 2
    WHOLE_CLASS = 3
    FORMATIVE = 101
    SUMMATIVE = 102
    COMPREHENSIVE = 103
    CONTEXT_CHOICES = (
        (NONE, _("None")),
        (INDIVIDUAL, _("Individual Work")),
        (GROUPS, _("Work in Groups")),
        (WHOLE_CLASS, _("Whole Class")),
        (FORMATIVE, _("Formative")),
        (SUMMATIVE, _("Summative")),
        (COMPREHENSIVE, _("Comprehensive")),
    )
    context_classification = models.PositiveIntegerField(
        choices=CONTEXT_CHOICES, default=0
    )
    GATHER_INFO = 1
    DISCUSS = 2
    PROBLEM_SOLVE = 3
    ANALYZE = 4
    ASSESS_PEERS = 5
    DEBATE = 6
    GAME_ROLEPLAY = 7
    CREATE_DESIGN = 8
    REVISE = 9
    READ = 10
    WRITE = 11
    PRESENT = 12
    EXPERIMENT = 13
    QUIZ_TEST = 14
    INSTRUCTOR_RESOURCE_CURATION = 15
    INSTRUCTOR_ORCHESTRATION = 16
    INSTRUCTOR_EVALUATION = 17
    OTHER = 18
    JIGSAW = 101
    PEER_INSTRUCTION = 102
    CASE_STUDIES = 103
    GALLERY_WALK = 104
    REFLECTIVE_WRITING = 105
    TWO_STAGE_EXAM = 106
    TOOLKIT = 107
    ONE_MINUTE_PAPER = 108
    DISTRIBUTED_PROBLEM_SOLVING = 109
    PEER_ASSESSMENT = 110
    TASK_CHOICES = (
        (NONE, _("None")),
        (GATHER_INFO, _("Gather Information")),
        (DISCUSS, _("Discuss")),
        (PROBLEM_SOLVE, _("Problem Solve")),
        (ANALYZE, _("Analyze")),
        (ASSESS_PEERS, _("Assess/Review Peers")),
        (DEBATE, _("Debate")),
        (GAME_ROLEPLAY, _("Game/Roleplay")),
        (CREATE_DESIGN, _("Create/Design")),
        (REVISE, _("Revise/Improve")),
        (READ, _("Read")),
        (WRITE, _("Write")),
        (PRESENT, _("Present")),
        (EXPERIMENT, _("Experiment/Inquiry")),
        (QUIZ_TEST, _("Quiz/Test")),
        (INSTRUCTOR_RESOURCE_CURATION, _("Instructor Resource Curation")),
        (INSTRUCTOR_ORCHESTRATION, _("Instructor Orchestration")),
        (INSTRUCTOR_EVALUATION, _("Instructor Evaluation")),
        (OTHER, _("Other")),
        (JIGSAW, _("Jigsaw")),
        (PEER_INSTRUCTION, _("Peer Instruction")),
        (CASE_STUDIES, _("Case Studies")),
        (GALLERY_WALK, _("Gallery Walk")),
        (REFLECTIVE_WRITING, _("Reflective Writing")),
        (TWO_STAGE_EXAM, _("Two-Stage Exam")),
        (TOOLKIT, _("Toolkit")),
        (ONE_MINUTE_PAPER, _("One Minute Paper")),
        (DISTRIBUTED_PROBLEM_SOLVING, _("Distributed Problem Solving")),
        (PEER_ASSESSMENT, _("Peer Assessment")),
    )
    task_classification = models.PositiveIntegerField(
        choices=TASK_CHOICES, default=0
    )
    ACTIVITY_NODE = 0
    COURSE_NODE = 1
    PROGRAM_NODE = 2
    NODE_TYPES = (
        (ACTIVITY_NODE, _("Activity Node")),
        (COURSE_NODE, _("Course Node")),
        (PROGRAM_NODE, _("Program Node")),
    )
    node_type = models.PositiveIntegerField(choices=NODE_TYPES, default=0)

    NO_UNITS = 0
    SECONDS = 1
    MINUTES = 2
    HOURS = 3
    DAYS = 4
    WEEKS = 5
    MONTHS = 6
    YEARS = 7
    CREDITS = 8
    UNIT_CHOICES = (
        (NO_UNITS, ""),
        (SECONDS, _("seconds")),
        (MINUTES, _("minutes")),
        (HOURS, _("hours")),
        (DAYS, _("days")),
        (WEEKS, _("weeks")),
        (MONTHS, _("months")),
        (YEARS, _("yrs")),
        (CREDITS, _("credits")),
    )

    # note: use charfield because some users like to put in ranges (i.e. 10-15 minutes)
    time_required = models.CharField(max_length=30, null=True, blank=True)
    time_units = models.PositiveIntegerField(default=0, choices=UNIT_CHOICES)

    ponderation_theory = models.PositiveIntegerField(default=0, null=True)
    ponderation_practical = models.PositiveIntegerField(default=0, null=True)
    ponderation_individual = models.PositiveIntegerField(default=0, null=True)

    time_general_hours = models.PositiveIntegerField(default=0, null=True)
    time_specific_hours = models.PositiveIntegerField(default=0, null=True)

    represents_workflow = models.BooleanField(default=False)
    linked_workflow = models.ForeignKey(
        "Workflow",
        on_delete=models.SET_NULL,
        null=True,
        related_name="linked_nodes",
    )

    column = models.ForeignKey(
        "Column", on_delete=models.DO_NOTHING, null=True
    )

    hash = models.UUIDField(default=uuid.uuid4, editable=False, unique=True)

    outcomes = models.ManyToManyField(
        Outcome, through="OutcomeNode", blank=True
    )

    def get_permission_objects(self):
        return self.get_workflow().get_permission_objects()

    def get_workflow(self):
        return self.week_set.first().get_workflow()

    def __str__(self):
        if self.title is not None:
            return self.title
        else:
            return self.get_node_type_display()

    class Meta:
        verbose_name = _("Node")
        verbose_name_plural = _("Nodes")


class OutcomeNode(models.Model):
    node = models.ForeignKey(Node, on_delete=models.CASCADE)
    outcome = models.ForeignKey(Outcome, on_delete=models.CASCADE)
    added_on = models.DateTimeField(default=timezone.now)
    rank = models.PositiveIntegerField(default=0)
    degree = models.PositiveIntegerField(default=1)

    def get_permission_objects(self):
        return self.get_workflow().get_permission_objects()

    def get_workflow(self):
        return self.node.get_workflow()

    def get_top_outcome(self):
        return self.outcome.get_top_outcome()

    class Meta:
        verbose_name = _("Outcome-Node Link")
        verbose_name_plural = _("Outcome-Node Links")

    # Check to see if the parent has all its children the same, and add it if necessary
    def check_parent_outcomes(self):
        if self.outcome.parent_outcomes.count() > 0:
            parent_outcome = self.outcome.parent_outcomes.first()
            if (
                OutcomeNode.objects.filter(
                    outcome__in=parent_outcome.children.exclude(
                        deleted=True
                    ).values_list("id", flat=True),
                    degree=self.degree,
                    node=self.node,
                ).count()
                == parent_outcome.children.exclude(deleted=True).count()
            ):
                new_outcomenode = OutcomeNode.objects.create(
                    node=self.node, degree=self.degree, outcome=parent_outcome
                )
                return [
                    new_outcomenode
                ] + new_outcomenode.check_parent_outcomes()
            elif (
                OutcomeNode.objects.filter(
                    outcome=parent_outcome, node=self.node
                ).count()
                > 0
            ):
                new_outcomenode = OutcomeNode.objects.create(
                    node=self.node, degree=0, outcome=parent_outcome
                )
                return [
                    new_outcomenode
                ] + new_outcomenode.check_parent_outcomes()

        return []

    # Check to see if the children already exist, and if not, add them
    def check_child_outcomes(self):

        node = self.node
        outcome = self.outcome
        degree = self.degree
        # Get the descendants (all descendant outcomes that don't already have an outcomenode of this degree and node)
        descendants = get_descendant_outcomes(outcome).exclude(
            outcomenode__in=OutcomeNode.objects.filter(
                node=node, degree=degree
            )
        )

        # Delete the outcomenodes of any descendants that still have an outcomenode to this node (i.e. clear those of other degrees, we are using bulk create so they won't get automatically deleted)
        to_delete = OutcomeNode.objects.filter(
            outcome__in=descendants.values_list("pk", flat=True), node=node
        )
        to_delete.delete()
        # Create the new outcomenodes with bulk_create
        now = timezone.now()
        new_children = [
            OutcomeNode(degree=degree, node=node, outcome=x, added_on=now)
            for x in descendants
        ]

        OutcomeNode.objects.bulk_create(new_children)

        return list(OutcomeNode.objects.filter(added_on=now))


class Week(models.Model):
    deleted = models.BooleanField(default=False)
    deleted_on = models.DateTimeField(default=timezone.now)
    title = models.CharField(
        max_length=title_max_length, null=True, blank=True
    )
    description = models.TextField(null=True, blank=True)
    author = models.ForeignKey(User, on_delete=models.SET_NULL, null=True)
    created_on = models.DateTimeField(default=timezone.now)
    last_modified = models.DateTimeField(auto_now=True)
    default = models.BooleanField(default=False)
    parent_week = models.ForeignKey(
        "Week", on_delete=models.SET_NULL, null=True
    )
    is_original = models.BooleanField(default=True)
    is_strategy = models.BooleanField(default=False)
    original_strategy = models.ForeignKey(
        "Workflow", on_delete=models.SET_NULL, null=True
    )

    is_dropped = models.BooleanField(default=True)

    hash = models.UUIDField(default=uuid.uuid4, editable=False, unique=True)

    nodes = models.ManyToManyField(Node, through="NodeWeek", blank=True)

    comments = models.ManyToManyField(
        "Comment", blank=True, related_name="week"
    )

    NONE = 0
    JIGSAW = 1
    PEER_INSTRUCTION = 2
    CASE_STUDIES = 3
    GALLERY_WALK = 4
    REFLECTIVE_WRITING = 5
    TWO_STAGE_EXAM = 6
    TOOLKIT = 7
    ONE_MINUTE_PAPER = 8
    DISTRIBUTED_PROBLEM_SOLVING = 9
    PEER_ASSESSMENT = 10
    OTHER = 11
    STRATEGY_CHOICES = (
        (NONE, _("None")),
        (JIGSAW, _("Jigsaw")),
        (PEER_INSTRUCTION, _("Peer Instruction")),
        (CASE_STUDIES, _("Case Studies")),
        (GALLERY_WALK, _("Gallery Walk")),
        (REFLECTIVE_WRITING, _("Reflective Writing")),
        (TWO_STAGE_EXAM, _("Two-Stage Exam")),
        (TOOLKIT, _("Toolkit")),
        (ONE_MINUTE_PAPER, _("One Minute Paper")),
        (DISTRIBUTED_PROBLEM_SOLVING, _("Distributed Problem Solving")),
        (PEER_ASSESSMENT, _("Peer Assessment")),
        (OTHER, _("Other")),
    )
    strategy_classification = models.PositiveIntegerField(
        choices=STRATEGY_CHOICES, default=0
    )

    PART = 0
    WEEK = 1
    TERM = 2
    WEEK_TYPES = ((PART, _("Part")), (WEEK, _("Week")), (TERM, _("Term")))
    week_type = models.PositiveIntegerField(choices=WEEK_TYPES, default=0)

    def __str__(self):
        return self.get_week_type_display()

    def get_permission_objects(self):
        return self.get_workflow().get_permission_objects()

    def get_workflow(self):
        return self.workflow_set.first()

    class Meta:
        verbose_name = _("Week")
        verbose_name_plural = _("Weeks")


class NodeWeek(models.Model):
    week = models.ForeignKey(Week, on_delete=models.CASCADE)
    node = models.ForeignKey(Node, on_delete=models.CASCADE)
    added_on = models.DateTimeField(default=timezone.now)
    rank = models.PositiveIntegerField(default=0)

    def get_workflow(self):
        return self.week.get_workflow()

    class Meta:
        verbose_name = _("Node-Week Link")
        verbose_name_plural = _("Node-Week Links")


class Workflow(models.Model):
    objects = InheritanceManager()

    author = models.ForeignKey(
        User,
        related_name="authored_workflows",
        on_delete=models.SET_NULL,
        null=True,
    )

    @property
    def importing(self):
        return cache.get("workflow" + str(self.pk) + "importing", False)

    deleted = models.BooleanField(default=False)
    deleted_on = models.DateTimeField(default=timezone.now)

    title = models.CharField(
        max_length=title_max_length, null=True, blank=True
    )
    description = models.TextField(null=True, blank=True)
    code = models.CharField(max_length=title_max_length, null=True, blank=True)
    created_on = models.DateTimeField(default=timezone.now)
    last_modified = models.DateTimeField(auto_now=True)

    static = models.BooleanField(default=False)

    published = models.BooleanField(default=False)

    public_view = models.BooleanField(default=False)

    is_strategy = models.BooleanField(default=False)

    from_saltise = models.BooleanField(default=False)

    condensed = models.BooleanField(default=False)

    user_permissions = GenericRelation(
        "ObjectPermission", related_query_name="workflow"
    )
    favourited_by = GenericRelation("Favourite", related_query_name="workflow")
    notified_by = GenericRelation("Notification", related_query_name="workflow")

    parent_workflow = models.ForeignKey(
        "Workflow", on_delete=models.SET_NULL, null=True
    )
    is_original = models.BooleanField(default=True)

    hash = models.UUIDField(default=uuid.uuid4, editable=False, unique=True)

    disciplines = models.ManyToManyField("Discipline", blank=True)
    weeks = models.ManyToManyField(Week, through="WeekWorkflow", blank=True)

    columns = models.ManyToManyField(
        Column, through="ColumnWorkflow", blank=True
    )

    outcomes = models.ManyToManyField(
        Outcome, through="OutcomeWorkflow", blank=True
    )

    OUTCOMES_NORMAL = 0
    OUTCOMES_ADVANCED = 1
    OUTCOME_TYPES = (
        (OUTCOMES_NORMAL, _("Normal")),
        (OUTCOMES_ADVANCED, _("Advanced")),
    )
    outcomes_type = models.PositiveIntegerField(
        choices=OUTCOME_TYPES, default=0
    )

    OUTCOME_SORT_WEEK = 0
    OUTCOME_SORT_COLUMN = 1
    OUTCOME_SORT_TASK = 2
    OUTCOME_SORT_CONTEXT = 3
    OUTCOME_SORTS = (
        (OUTCOME_SORT_WEEK, _("Time")),
        (OUTCOME_SORT_COLUMN, _("Category")),
        (OUTCOME_SORT_TASK, _("Task")),
        (OUTCOME_SORT_CONTEXT, _("Context")),
    )
    outcomes_sort = models.PositiveIntegerField(
        choices=OUTCOME_SORTS, default=0
    )

    NO_UNITS = 0
    SECONDS = 1
    MINUTES = 2
    HOURS = 3
    DAYS = 4
    WEEKS = 5
    MONTHS = 6
    YEARS = 7
    CREDITS = 8
    UNIT_CHOICES = (
        (NO_UNITS, ""),
        (SECONDS, _("seconds")),
        (MINUTES, _("minutes")),
        (HOURS, _("hours")),
        (DAYS, _("days")),
        (WEEKS, _("weeks")),
        (MONTHS, _("months")),
        (YEARS, _("yrs")),
        (CREDITS, _("credits")),
    )

    # note: use charfield because some users like to put in ranges (i.e. 10-15 minutes)
    time_required = models.CharField(max_length=30, null=True, blank=True)
    time_units = models.PositiveIntegerField(default=0, choices=UNIT_CHOICES)

    ponderation_theory = models.PositiveIntegerField(default=0, null=True)
    ponderation_practical = models.PositiveIntegerField(default=0, null=True)
    ponderation_individual = models.PositiveIntegerField(default=0, null=True)

    time_general_hours = models.PositiveIntegerField(default=0, null=True)
    time_specific_hours = models.PositiveIntegerField(default=0, null=True)

    edit_count = models.PositiveIntegerField(default=0, null=False)

    SUBCLASSES = ["activity", "course", "program"]

    @property
    def type(self):
        for subclass in self.SUBCLASSES:
            try:
                return getattr(self, subclass).type
            except AttributeError:
                pass
        return "workflow"

    def get_project(self):
        return self.project_set.first()

    def get_workflow(self):
        return Workflow.objects.get(pk=self.pk)

    def get_permission_objects(self):
        return [self.get_workflow()]

    def get_live_project(self):
        try:
            liveproject = self.get_project().liveproject
        except AttributeError:
            liveproject = None
        return liveproject

    def get_subclass(self):
        subclass = self
        try:
            subclass = self.activity
        except AttributeError:
            pass
        try:
            subclass = self.course
        except AttributeError:
            pass
        try:
            subclass = self.program
        except AttributeError:
            pass
        return subclass

    def get_all_outcome_ids(self):
        ids = []
        for outcome in self.outcomes.all():
            outcome.get_all_outcome_ids(ids)
        return ids

    def __str__(self):
        if self.title is not None and self.title != "":
            return self.title
        else:
            return self.type

    class Meta:
        verbose_name = _("Workflow")
        verbose_name_plural = _("Workflows")


class Activity(Workflow):

    DEFAULT_CUSTOM_COLUMN = 0
    DEFAULT_COLUMNS = [1, 2, 3, 4]
    WORKFLOW_TYPE = 0

    @property
    def type(self):
        return "activity"

    def __str__(self):
        if self.title is not None and self.title != "":
            return self.title
        else:
            return self.type

    class Meta:
        verbose_name = _("Activity")
        verbose_name_plural = _("Activities")


class Course(Workflow):

    DEFAULT_CUSTOM_COLUMN = 10
    DEFAULT_COLUMNS = [11, 12, 13, 14]
    WORKFLOW_TYPE = 1

    @property
    def type(self):
        return "course"

    def __str__(self):
        if self.title is not None and self.title != "":
            return self.title
        else:
            return self.type

    class Meta:
        verbose_name = _("Course")
        verbose_name_plural = _("Courses")


class Program(Workflow):

    DEFAULT_CUSTOM_COLUMN = 20
    DEFAULT_COLUMNS = [20, 20, 20]
    WORKFLOW_TYPE = 2

    @property
    def type(self):
        return "program"

    def __str__(self):
        if self.title is not None and self.title != "":
            return self.title
        else:
            return self.type

    class Meta:
        verbose_name = _("Program")
        verbose_name_plural = _("Programs")


class ColumnWorkflow(models.Model):
    workflow = models.ForeignKey(Workflow, on_delete=models.CASCADE)
    column = models.ForeignKey(Column, on_delete=models.CASCADE)
    added_on = models.DateTimeField(default=timezone.now)
    rank = models.PositiveIntegerField(default=0)

    def get_workflow(self):
        return self.workflow

    def get_permission_objects(self):
        return self.get_workflow().get_permission_objects()

    class Meta:
        verbose_name = _("Column-Workflow Link")
        verbose_name_plural = _("Column-Workflow Links")


class WeekWorkflow(models.Model):
    workflow = models.ForeignKey(Workflow, on_delete=models.CASCADE)
    week = models.ForeignKey(Week, on_delete=models.CASCADE)
    added_on = models.DateTimeField(default=timezone.now)
    rank = models.PositiveIntegerField(default=0)

    def get_display_rank(self):
        if self.week.deleted:
            return -1
        return list(
            WeekWorkflow.objects.filter(
                workflow=self.workflow, week__deleted=False
            ).order_by("rank")
        ).index(self)

    def get_workflow(self):
        return self.workflow

    def get_permission_objects(self):
        return self.get_workflow().get_permission_objects()

    class Meta:
        verbose_name = _("Week-Workflow Link")
        verbose_name_plural = _("Week-Workflow Links")


class Discipline(models.Model):
    title = models.CharField(
        _("Discipline name"),
        unique=True,
        max_length=100,
        help_text=_("Enter the name of a new discipline."),
    )

    def __str__(self):
        return self.title

    class Meta:
        verbose_name = _("Discipline")
        verbose_name_plural = _("Disciplines")


workflow_choices = [
    "activity",
    "course",
    "program",
]


class Favourite(models.Model):
    content_choices = {"model__in": ["project", "workflow"]}
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, limit_choices_to=content_choices
    )
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey("content_type", "object_id")

    def save(self, *args, **kwargs):
        if self.content_object.type in workflow_choices:
            self.content_type = ContentType.objects.get_for_model(Workflow)
        super().save(*args, **kwargs)


class Comment(models.Model):
    user = models.ForeignKey(User, on_delete=models.SET_NULL, null=True)
    text = models.TextField(
        blank=False,
    )
    created_on = models.DateTimeField(default=timezone.now)


class ObjectPermission(models.Model):
    content_choices = {"model__in": ["project", "workflow"]}
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, limit_choices_to=content_choices
    )
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey("content_type", "object_id")

    PERMISSION_NONE = 0
    PERMISSION_VIEW = 1
    PERMISSION_EDIT = 2
    PERMISSION_COMMENT = 3
    PERMISSION_STUDENT = 4
    PERMISSION_CHOICES = (
        (PERMISSION_NONE, _("None")),
        (PERMISSION_VIEW, _("View")),
        (PERMISSION_EDIT, _("Edit")),
        (PERMISSION_COMMENT, _("Comment")),
        (PERMISSION_STUDENT, _("Student")),
    )
    permission_type = models.PositiveIntegerField(
        choices=PERMISSION_CHOICES, default=PERMISSION_NONE
    )

    last_viewed = models.DateTimeField(default=timezone.now)

    def update_last_viewed(user, view_object):
        ObjectPermission.objects.filter(
            user=user,
            content_type=ContentType.objects.get_for_model(view_object),
            object_id=view_object.id,
        ).update(last_viewed=timezone.now())

    def save(self, *args, **kwargs):
        if self.content_object.type in workflow_choices:
            self.content_type = ContentType.objects.get_for_model(Workflow)
        super().save(*args, **kwargs)


class UpdateNotification(models.Model):
    title = models.CharField(
        max_length=title_max_length, null=True, blank=True
    )
    created_on = models.DateTimeField(default=timezone.now)

    def __str__(self):
        if self.title is not None and self.title != "":
            return self.title
        else:
            return "Untitled Update Notification"


"""
Live Project Models
"""


def default_start_date():
    return timezone.now().replace(
        second=0, microsecond=0, minute=0, hour=0
    ) + timezone.timedelta(weeks=1)


def default_due_date():
    return timezone.now().replace(
        second=0, microsecond=0, minute=0, hour=0
    ) + timezone.timedelta(weeks=2)


class LiveProject(models.Model):

    created_on = models.DateTimeField(default=timezone.now)

    project = models.OneToOneField(
        Project, on_delete=models.CASCADE, primary_key=True
    )

    # Whether students are able to check tasks as complete themselves or
    # must have the instructor mark them as complete
    default_self_reporting = models.BooleanField(default=True)
    # Whether newly created assignments are assigned to all by default
    default_assign_to_all = models.BooleanField(default=True)
    # Whether it is enough for a single assigned user to complete the task,
    # or (when True) when any user completes the task it becomes complete for all users
    default_single_completion = models.BooleanField(default=False)
    # whether workflows are always all visible
    default_all_workflows_visible = models.BooleanField(default=False)
    # These workflows are always visible to all students
    visible_workflows = models.ManyToManyField(Workflow, blank=True)

    def get_permission_objects(self):
        return [self]

    def get_live_project(self):
        return self

    @property
    def type(self):
        return "liveproject"

    @property
    def last_modified(self):
        return self.project.last_modified

    @property
    def id(self):
        return self.pk

    @property
    def author(self):
        return self.project.author

    @property
    def title(self):
        return self.project.title

    @property
    def description(self):
        return self.project.description


class LiveProjectUser(models.Model):
    liveproject = models.ForeignKey(LiveProject, on_delete=models.CASCADE)
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    ROLE_NONE = 0
    ROLE_STUDENT = 1
    ROLE_TEACHER = 2
    ROLE_CHOICES = (
        (ROLE_NONE, _("None")),
        (ROLE_STUDENT, _("Student")),
        (ROLE_TEACHER, _("Instructor")),
    )
    role_type = models.PositiveIntegerField(
        choices=ROLE_CHOICES, default=ROLE_NONE
    )


class LiveAssignment(models.Model):
    author = models.ForeignKey(User, on_delete=models.SET_NULL, null=True)
    liveproject = models.ForeignKey(LiveProject, on_delete=models.CASCADE)
    self_reporting = models.BooleanField(default=True)
    single_completion = models.BooleanField(default=False)
    task = models.ForeignKey(Node, null=True, on_delete=models.SET_NULL)
    start_date = models.DateTimeField(default=default_start_date)
    end_date = models.DateTimeField(default=default_due_date)
    created_on = models.DateTimeField(default=timezone.now)

    @property
    def title(self):
        if self.task is not None:
            return self.task.title

    def get_live_project(self):
        return self.liveproject

    def get_permission_objects(self):
        return [self.liveproject]


class UserAssignment(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    assignment = models.ForeignKey(LiveAssignment, on_delete=models.CASCADE)
    completed = models.BooleanField(default=False)
    completed_on = models.DateTimeField(null=True)

    def get_live_project(self):
        return self.assignment.liveproject

    def get_permission_objects(self):
        return [self.assignment.liveproject]


"""
User Models
"""


class CourseFlowUser(models.Model):
    user = models.OneToOneField(
        User,
        on_delete=models.CASCADE,
        primary_key=True,
        related_name="courseflow_user",
    )
    first_name = models.CharField(
        max_length=title_max_length,
        null=True,
        blank=True,
    )
    last_name = models.CharField(
        max_length=title_max_length, null=True, blank=True
    )

    # Whether the user wants to receive notifications
    notifications = models.BooleanField(
        default=False,
        help_text="Check this box if you would like to receive emails from us about updates to CourseFlow.",
    )

    # Whether the user has had the opportunity to choose whether they receive notifications
    notifications_active = models.BooleanField(default=False)

    def ensure_user(user):
        courseflow_user = CourseFlowUser.objects.filter(user=user).first()
        if courseflow_user is None:
            courseflow_user = CourseFlowUser.objects.create(
                first_name=user.first_name,
                last_name=user.last_name,
                user=user,
            )
        return courseflow_user


class Notification(models.Model):
    class Meta:
        ordering = ["-created_on"]

    content_choices = {"model__in": ["project", "workflow"]}
    user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name="notifications",
    )
    source_user = models.ForeignKey(
        User,
        null=True,
        on_delete=models.SET_NULL,
        related_name="sent_notifications",
    )
    TYPE_SHARED = 0
    TYPE_COMMENT = 1
    TYPE_CHOICES = (
        (TYPE_SHARED, _("Shared")),
        (TYPE_COMMENT, _("Comment")),
    )
    notification_type = models.PositiveIntegerField(
        choices=TYPE_CHOICES, default=TYPE_SHARED
    )
    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, limit_choices_to=content_choices
    )
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey("content_type", "object_id")
    text = models.TextField(
        blank=True,
        null=True,
    )

    created_on = models.DateTimeField(default=timezone.now)
    is_unread = models.BooleanField(default=True)

    comment = models.ForeignKey(
        Comment,
        on_delete=models.CASCADE,
        related_name="notifications",
        null=True,
    )


"""
Other receivers
"""


def get_allowed_parent_outcomes(workflow, **kwargs):
    exclude_node = kwargs.get("exclude_node", None)
    exclude_outcomenode = kwargs.get("exclude_outcomenode", None)
    parent_outcomenodes = []
    for parent_node in workflow.linked_nodes.exclude(id=exclude_node):
        parent_outcomenodes += parent_node.outcomenode_set.exclude(
            id=exclude_outcomenode
        )
    parent_outcomes = [ocn.outcome for ocn in parent_outcomenodes]
    return parent_outcomes


# @receiver(pre_delete, sender=OutcomeNode)
# def remove_horizontal_outcome_links_on_outcomenode_delete(
#    sender, instance, **kwargs
# ):
#    workflow = instance.node.linked_workflow
#    if workflow is not None:
#        linked_outcomes = list(workflow.outcomes.all())
#        parent_outcomes = get_allowed_parent_outcomes(
#            workflow, exclude_outcomenode=instance.pk
#        )
#        OutcomeHorizontalLink.objects.filter(
#            outcome__in=linked_outcomes
#        ).exclude(parent_outcome__in=parent_outcomes).delete()


# @receiver(pre_save, sender=Node)
# def remove_horizontal_outcome_links_on_node_unlink(sender, instance, **kwargs):
#    if instance.pk is None:
#        return
#    old_workflow = Node.objects.get(id=instance.pk).linked_workflow
#    new_workflow = instance.linked_workflow
#    if old_workflow is not None and (
#        new_workflow is None or new_workflow.pk != old_workflow.pk
#    ):
#        linked_outcomes = list(old_workflow.outcomes.all())
#        parent_outcomes = get_allowed_parent_outcomes(
#            old_workflow, exclude_node=instance.pk
#        )
#        OutcomeHorizontalLink.objects.filter(
#            outcome__in=linked_outcomes
#        ).exclude(parent_outcome__in=parent_outcomes).delete()


@receiver(pre_delete, sender=Project)
def delete_project_objects(sender, instance, **kwargs):

    # Pick up all non-linking instances pks
    nodes = list(
        Node.objects.filter(week__workflow__project=instance).values_list(
            "pk", flat=True
        )
    )
    weeks = list(
        Week.objects.filter(workflow__project=instance).values_list(
            "pk", flat=True
        )
    )
    columns = list(
        Column.objects.filter(workflow__project=instance).values_list(
            "pk", flat=True
        )
    )
    outcomes = list(
        Outcome.objects.filter(
            Q(workflow__project=instance)
            | Q(parent_outcomes__workflow__project=instance)
            | Q(parent_outcomes__parent_outcomes__workflow__project=instance)
        ).values_list("pk", flat=True)
    )
    workflows = list(
        Workflow.objects.filter(project=instance).values_list("pk", flat=True)
    )
    comments = Comment.objects.filter(
        Q(node__week__workflow__project=instance)
        | Q(outcome__in=outcomes)
        | Q(column__workflow__project=instance)
        | Q(week__workflow__project=instance)
    )
    comments.delete()

    # Delete all links. These should be deleted before non-linking instances because this way we prevent a lot of cascades. Order matters here; we want to go from top to bottom or else we will break the links we need in order to find the next step
    outcomenodes = OutcomeNode.objects.filter(
        node__week__workflow__project=instance
    )
    outcomenodes._raw_delete(outcomenodes.db)
    nodelinks = NodeLink.objects.filter(
        source_node__week__workflow__project=instance
    )
    nodelinks._raw_delete(nodelinks.db)
    outcomehorizontallinks = OutcomeHorizontalLink.objects.filter(
        Q(outcome__workflow__project=instance)
        | Q(outcome__parent_outcomes__workflow__project=instance)
        | Q(
            outcome__parent_outcomes__parent_outcomes__workflow__project=instance
        )
    )
    outcomehorizontallinks._raw_delete(outcomehorizontallinks.db)
    nodeweeks = NodeWeek.objects.filter(week__workflow__project=instance)
    nodeweeks._raw_delete(nodeweeks.db)
    weekworkflows = WeekWorkflow.objects.filter(workflow__project=instance)
    weekworkflows._raw_delete(weekworkflows.db)
    columnworkflows = ColumnWorkflow.objects.filter(workflow__project=instance)
    columnworkflows._raw_delete(columnworkflows.db)
    outcomeoutcomes = OutcomeOutcome.objects.filter(
        Q(parent__workflow__project=instance)
        | Q(parent__parent_outcomes__workflow__project=instance)
    )
    outcomeoutcomes._raw_delete(outcomeoutcomes.db)
    outcomeworkflows = OutcomeWorkflow.objects.filter(
        workflow__project=instance
    )
    outcomeworkflows._raw_delete(outcomeworkflows.db)
    workflowprojects = WorkflowProject.objects.filter(project=instance)
    workflowprojects._raw_delete(workflowprojects.db)

    # remove all FKs pointing to our objects from outside project. The raw deletes don't cascade, so we will get integrity errors if we fail to do this
    #    workflow_subclasses = [workflow.get_subclass() for workflow in workflows]
    #    activities = filter(lambda x: x.type == "activity", workflow_subclasses)
    #    courses = filter(lambda x: x.type == "course", workflow_subclasses)
    #    programs = filter(lambda x: x.type == "program", workflow_subclasses)
    objectpermissions = ObjectPermission.objects.filter(
        Q(workflow__in=workflows) | Q(project=instance)
    )
    favourites = Favourite.objects.filter(
        Q(workflow__in=workflows) | Q(project=instance)
    )
    notifications = Notification.objects.filter(
        Q(workflow__in=workflows) | Q(project=instance)
    )
    Node.objects.filter(parent_node__in=nodes).update(parent_node=None)
    Node.objects.filter(linked_workflow__in=workflows).update(
        linked_workflow=None
    )
    Week.objects.filter(parent_week__in=weeks).update(parent_week=None)
    Week.objects.filter(original_strategy__in=workflows).update(
        original_strategy=None
    )
    Column.objects.filter(parent_column__in=columns).update(parent_column=None)
    Workflow.objects.filter(parent_workflow__in=workflows).update(
        parent_workflow=None
    )
    Outcome.objects.filter(parent_outcome__in=outcomes).update(
        parent_outcome=None
    )

    # Delete nonlinking instances
    nodes = Node.objects.filter(pk__in=nodes)
    nodes._raw_delete(nodes.db)
    weeks = Week.objects.filter(pk__in=weeks)
    weeks._raw_delete(weeks.db)
    columns = Column.objects.filter(pk__in=columns)
    columns._raw_delete(columns.db)
    outcomes = Outcome.objects.filter(pk__in=outcomes)
    outcomes._raw_delete(outcomes.db)
    objectpermissions._raw_delete(objectpermissions.db)
    favourites._raw_delete(favourites.db)
    notifications._raw_delete(notifications.db)
    activities = Activity.objects.filter(pk__in=workflows)
    activities._raw_delete(activities.db)
    courses = Course.objects.filter(pk__in=workflows)
    courses._raw_delete(courses.db)
    programs = Program.objects.filter(pk__in=workflows)
    programs._raw_delete(programs.db)
    workflows = Workflow.objects.filter(pk__in=workflows)
    workflows.delete()
    # raw delete was presenting issues with the disciplines for some reason
    # Given that most usage is soft delete, might as well just .delete()
    # workflows._raw_delete(workflows.db)


@receiver(pre_delete, sender=Workflow)
def delete_workflow_objects(sender, instance, **kwargs):

    # Pick up all non-linking instances pks
    nodes = list(
        Node.objects.filter(week__workflow=instance).values_list(
            "pk", flat=True
        )
    )
    weeks = list(
        Week.objects.filter(workflow=instance).values_list("pk", flat=True)
    )
    columns = list(
        Column.objects.filter(workflow=instance).values_list("pk", flat=True)
    )
    outcomes = list(
        Outcome.objects.filter(
            Q(workflow=instance)
            | Q(parent_outcomes__workflow=instance)
            | Q(parent_outcomes__parent_outcomes__workflow=instance)
        ).values_list("pk", flat=True)
    )

    # Delete all comments.
    comments = Comment.objects.filter(
        Q(node__week__workflow=instance)
        | Q(outcome__in=outcomes)
        | Q(column__workflow=instance)
        | Q(week__workflow=instance)
    )
    comments.delete()

    # Delete all links. These should be deleted before non-linking instances because this way we prevent a lot of cascades. Order matters here; we want to go from top to bottom or else we will break the links we need in order to find the next step
    outcomenodes = OutcomeNode.objects.filter(node__week__workflow=instance)
    outcomenodes._raw_delete(outcomenodes.db)
    nodelinks = NodeLink.objects.filter(source_node__week__workflow=instance)
    nodelinks._raw_delete(nodelinks.db)
    outcomehorizontallinks = OutcomeHorizontalLink.objects.filter(
        Q(outcome__workflow=instance)
        | Q(outcome__parent_outcomes__workflow=instance)
        | Q(outcome__parent_outcomes__parent_outcomes__workflow=instance)
        | Q(parent_outcome__workflow=instance)
        | Q(parent_outcome__parent_outcomes__workflow=instance)
        | Q(
            parent_outcome__parent_outcomes__parent_outcomes__workflow=instance
        )
    )
    outcomehorizontallinks._raw_delete(outcomehorizontallinks.db)
    nodeweeks = NodeWeek.objects.filter(week__workflow=instance)
    nodeweeks._raw_delete(nodeweeks.db)
    weekworkflows = WeekWorkflow.objects.filter(workflow=instance)
    weekworkflows._raw_delete(weekworkflows.db)
    columnworkflows = ColumnWorkflow.objects.filter(workflow=instance)
    columnworkflows._raw_delete(columnworkflows.db)
    outcomeoutcomes = OutcomeOutcome.objects.filter(
        Q(parent__workflow=instance)
        | Q(parent__parent_outcomes__workflow=instance)
    )
    outcomeoutcomes._raw_delete(outcomeoutcomes.db)
    outcomeworkflows = OutcomeWorkflow.objects.filter(workflow=instance)
    outcomeworkflows._raw_delete(outcomeworkflows.db)

    # remove all FKs pointing to our objects from outside project. The raw deletes don't cascade, so we will get integrity errors if we fail to do this
    Node.objects.filter(parent_node__in=nodes).update(parent_node=None)
    Week.objects.filter(parent_week__in=weeks).update(parent_week=None)
    Column.objects.filter(parent_column__in=columns).update(parent_column=None)
    Outcome.objects.filter(parent_outcome__in=outcomes).update(
        parent_outcome=None
    )
    favourites = Favourite.objects.filter(
        workflow=instance
    )
    notifications = Notification.objects.filter(
        workflow=instance
    )

    # Delete nonlinking instances
    nodes = Node.objects.filter(pk__in=nodes)
    nodes._raw_delete(nodes.db)
    weeks = Week.objects.filter(pk__in=weeks)
    weeks._raw_delete(weeks.db)
    columns = Column.objects.filter(pk__in=columns)
    columns._raw_delete(columns.db)
    outcomes = Outcome.objects.filter(pk__in=outcomes)
    outcomes._raw_delete(outcomes.db)
    favourites._raw_delete(favourites.db)
    notifications._raw_delete(notifications.db)


@receiver(pre_delete, sender=Week)
def delete_week_objects(sender, instance, **kwargs):
    instance.nodes.all().delete()


@receiver(pre_delete, sender=Node)
def delete_node_objects(sender, instance, **kwargs):
    instance.outgoing_links.all().delete()
    instance.incoming_links.all().delete()
    if instance.linked_workflow is not None:
        instance.linked_workflow = None
        instance.save()


@receiver(pre_delete, sender=Outcome)
def delete_outcome_objects(sender, instance, **kwargs):
    instance.children.all().delete()


@receiver(pre_delete, sender=Column)
def move_nodes(sender, instance, **kwargs):
    columnworkflow = instance.columnworkflow_set.first()
    if columnworkflow is None:
        print("This column has no columnworkflow, probably orphaned")
        return
    workflow = columnworkflow.workflow

    other_columns = (
        workflow.columnworkflow_set.all()
        .order_by("rank")
        .exclude(column=instance)
    )
    if other_columns.count() > 0:
        new_column = other_columns.first().column
        for node in Node.objects.filter(column=instance):
            node.column = new_column
            node.save()
    else:
        print("couldn't find a column")


"""
Reorder Receivers
"""


@receiver(pre_delete, sender=NodeWeek)
def reorder_for_deleted_node_week(sender, instance, **kwargs):
    for out_of_order_link in NodeWeek.objects.filter(
        week=instance.week, rank__gt=instance.rank
    ):
        out_of_order_link.rank -= 1
        out_of_order_link.save()


@receiver(pre_delete, sender=WeekWorkflow)
def reorder_for_deleted_week_workflow(sender, instance, **kwargs):
    for out_of_order_link in WeekWorkflow.objects.filter(
        workflow=instance.workflow, rank__gt=instance.rank
    ):
        out_of_order_link.rank -= 1
        out_of_order_link.save()


@receiver(pre_delete, sender=ColumnWorkflow)
def reorder_for_deleted_column_workflow(sender, instance, **kwargs):
    for out_of_order_link in ColumnWorkflow.objects.filter(
        workflow=instance.workflow, rank__gt=instance.rank
    ):
        out_of_order_link.rank -= 1
        out_of_order_link.save()


@receiver(pre_delete, sender=OutcomeWorkflow)
def reorder_for_deleted_outcome_workflow(sender, instance, **kwargs):
    for out_of_order_link in OutcomeWorkflow.objects.filter(
        workflow=instance.workflow, rank__gt=instance.rank
    ):
        out_of_order_link.rank -= 1
        out_of_order_link.save()


@receiver(pre_delete, sender=OutcomeOutcome)
def reorder_for_deleted_outcome_outcome(sender, instance, **kwargs):
    for out_of_order_link in OutcomeOutcome.objects.filter(
        parent=instance.parent, rank__gt=instance.rank
    ):
        out_of_order_link.rank -= 1
        out_of_order_link.save()


#
# @receiver(pre_delete, sender=OutcomeNode)
# def reorder_for_deleted_outcome_node(sender, instance, **kwargs):
#    for out_of_order_link in OutcomeNode.objects.filter(
#        node=instance.node, rank__gt=instance.rank
#    ):
#        out_of_order_link.rank -= 1
#        out_of_order_link.save()

#
# @receiver(pre_delete, sender=OutcomeHorizontalLink)
# def reorder_for_deleted_outcome_horizontal_link(sender, instance, **kwargs):
#    for out_of_order_link in OutcomeNode.objects.filter(
#        outcome=instance.outcome, rank__gt=instance.rank
#    ):
#        out_of_order_link.rank -= 1
#        out_of_order_link.save()


@receiver(pre_save, sender=WorkflowProject)
def delete_existing_workflow_project(sender, instance, **kwargs):
    if instance.pk is None:
        WorkflowProject.objects.filter(workflow=instance.workflow).delete()
        if instance.rank < 0:
            instance.rank = 0
        new_parent_count = WorkflowProject.objects.filter(
            project=instance.project
        ).count()
        if instance.rank > new_parent_count:
            instance.rank = new_parent_count


@receiver(pre_save, sender=NodeWeek)
def delete_existing_node_week(sender, instance, **kwargs):
    if instance.pk is None:
        try:
            NodeWeek.objects.filter(node=instance.node).delete()
        except Exception as e:
            print(e)
        if instance.rank < 0:
            instance.rank = 0
        new_parent_count = NodeWeek.objects.filter(week=instance.week).count()
        if instance.rank > new_parent_count:
            instance.rank = new_parent_count


@receiver(post_save, sender=NodeWeek)
def reorder_for_created_node_week(sender, instance, created, **kwargs):
    if created:
        for out_of_order_link in NodeWeek.objects.filter(
            week=instance.week, rank__gte=instance.rank
        ).exclude(node=instance.node):
            out_of_order_link.rank += 1
            out_of_order_link.save()


@receiver(pre_save, sender=WeekWorkflow)
def delete_existing_week_workflow(sender, instance, **kwargs):
    if instance.pk is None:
        WeekWorkflow.objects.filter(week=instance.week).delete()
        if instance.rank < 0:
            instance.rank = 0
        new_parent_count = WeekWorkflow.objects.filter(
            workflow=instance.workflow
        ).count()
        if instance.rank > new_parent_count:
            instance.rank = new_parent_count


@receiver(post_save, sender=WeekWorkflow)
def reorder_for_created_week_workflow(sender, instance, created, **kwargs):
    if created:
        for out_of_order_link in WeekWorkflow.objects.filter(
            workflow=instance.workflow, rank__gte=instance.rank
        ).exclude(week=instance.week):
            out_of_order_link.rank += 1
            out_of_order_link.save()


@receiver(pre_save, sender=ColumnWorkflow)
def delete_existing_column_workflow(sender, instance, **kwargs):
    if instance.pk is None:
        ColumnWorkflow.objects.filter(column=instance.column).delete()
        if instance.rank < 0:
            instance.rank = 0
        new_parent_count = ColumnWorkflow.objects.filter(
            workflow=instance.workflow
        ).count()
        if instance.rank > new_parent_count:
            instance.rank = new_parent_count


@receiver(post_save, sender=ColumnWorkflow)
def reorder_for_created_column_workflow(sender, instance, created, **kwargs):
    if created:
        for out_of_order_link in ColumnWorkflow.objects.filter(
            workflow=instance.workflow, rank__gte=instance.rank
        ).exclude(column=instance.column):
            out_of_order_link.rank += 1
            out_of_order_link.save()


@receiver(pre_save, sender=OutcomeWorkflow)
def delete_existing_outcome_workflow(sender, instance, **kwargs):
    if instance.pk is None:
        OutcomeWorkflow.objects.filter(outcome=instance.outcome).delete()
        if instance.rank < 0:
            instance.rank = 0
        new_parent_count = OutcomeWorkflow.objects.filter(
            workflow=instance.workflow
        ).count()
        if instance.rank > new_parent_count:
            instance.rank = new_parent_count


@receiver(post_save, sender=OutcomeWorkflow)
def reorder_for_created_outcome_workflow(sender, instance, created, **kwargs):
    if created:
        for out_of_order_link in OutcomeWorkflow.objects.filter(
            workflow=instance.workflow, rank__gte=instance.rank
        ).exclude(outcome=instance.outcome):
            out_of_order_link.rank += 1
            out_of_order_link.save()


@receiver(pre_save, sender=OutcomeOutcome)
def delete_existing_outcome_outcome(sender, instance, **kwargs):
    if instance.pk is None:
        OutcomeOutcome.objects.filter(child=instance.child).delete()
        if instance.rank < 0:
            instance.rank = 0
        new_parent_count = OutcomeOutcome.objects.filter(
            parent=instance.parent
        ).count()
        if instance.rank > new_parent_count:
            instance.rank = new_parent_count


@receiver(post_save, sender=OutcomeOutcome)
def reorder_for_created_outcome_outcome(sender, instance, created, **kwargs):
    if created:
        for out_of_order_link in OutcomeOutcome.objects.filter(
            parent=instance.parent, rank__gte=instance.rank
        ).exclude(child=instance.child):
            out_of_order_link.rank += 1
            out_of_order_link.save()


@receiver(pre_save, sender=OutcomeNode)
def delete_existing_outcome_node(sender, instance, **kwargs):
    if instance.pk is None:
        OutcomeNode.objects.filter(
            node=instance.node, outcome=instance.outcome
        ).delete()
        if instance.rank < 0:
            instance.rank = 0
        new_parent_count = OutcomeNode.objects.filter(
            node=instance.node
        ).count()
        if instance.rank > new_parent_count:
            instance.rank = new_parent_count


#
#
# @receiver(post_save, sender=OutcomeNode)
# def reorder_for_created_outcome_node(sender, instance, created, **kwargs):
#    if created:
#        for out_of_order_link in OutcomeNode.objects.filter(
#            node=instance.node, rank__gte=instance.rank
#        ).exclude(outcome=instance.outcome):
#            out_of_order_link.rank += 1
#            out_of_order_link.save()


@receiver(pre_save, sender=OutcomeHorizontalLink)
def delete_existing_horizontal_link(sender, instance, **kwargs):
    if instance.pk is None:
        OutcomeHorizontalLink.objects.filter(
            outcome=instance.outcome, parent_outcome=instance.parent_outcome
        ).delete()
        if instance.rank < 0:
            instance.rank = 0
        new_parent_count = OutcomeHorizontalLink.objects.filter(
            outcome=instance.outcome
        ).count()
        if instance.rank > new_parent_count:
            instance.rank = new_parent_count


# @receiver(post_save, sender=OutcomeHorizontalLink)
# def reorder_for_created_horizontal_outcome_link(
#    sender, instance, created, **kwargs
# ):
#    if created:
#        for out_of_order_link in OutcomeHorizontalLink.objects.filter(
#            outcome=instance.outcome, rank__gte=instance.rank
#        ).exclude(parent_outcome=instance.parent_outcome):
#            out_of_order_link.rank += 1
#            out_of_order_link.save()


"""
Default content creation receivers
"""


@receiver(post_save, sender=ObjectPermission)
def set_permissions_to_project_objects(sender, instance, created, **kwargs):
    if created:
        if instance.content_type == ContentType.objects.get_for_model(Project):
            if instance.permission_type == ObjectPermission.PERMISSION_STUDENT:
                liveproject = instance.content_object.liveproject
                if liveproject is not None:
                    if liveproject.default_all_workflows_visible:
                        workflows = instance.content_object.workflows.all()
                    else:
                        workflows = liveproject.visible_workflows.all()
            else:
                workflows = instance.content_object.workflows.all()
            for workflow in workflows:
                # If user already has edit or comment permissions and we are adding view, do not override
                if (
                    instance.permission_type
                    == ObjectPermission.PERMISSION_VIEW
                    and ObjectPermission.objects.filter(
                        user=instance.user,
                        content_type=ContentType.objects.get_for_model(
                            workflow
                        ),
                        object_id=workflow.id,
                        permission_type__in=[
                            ObjectPermission.PERMISSION_EDIT,
                            ObjectPermission.PERMISSION_COMMENT,
                        ],
                    ).count()
                    > 0
                ):
                    pass
                elif (
                    instance.permission_type
                    == ObjectPermission.PERMISSION_COMMENT
                    and ObjectPermission.objects.filter(
                        user=instance.user,
                        content_type=ContentType.objects.get_for_model(
                            workflow
                        ),
                        object_id=workflow.id,
                        permission_type__in=[ObjectPermission.PERMISSION_EDIT],
                    ).count()
                    > 0
                ):
                    pass
                else:
                    # If user is the owner, don't override their ownership
                    if workflow.author == instance.user:
                        if (
                            ObjectPermission.objects.filter(
                                workflow=workflow,
                                user=instance.user,
                                permission_type=ObjectPermission.PERMISSION_EDIT,
                            ).count()
                            == 0
                        ):
                            # Just in case the user has somehow lost their permission
                            ObjectPermission.objects.create(
                                user=instance.user,
                                content_object=workflow,
                                permission_type=ObjectPermission.PERMISSION_EDIT,
                            )
                    else:
                        ObjectPermission.objects.create(
                            user=instance.user,
                            content_object=workflow,
                            permission_type=instance.permission_type,
                        )


        elif instance.content_type == ContentType.objects.get_for_model(Workflow):
            workflow = instance.content_object
            project = workflow.get_project()
            if not project is None:
                if ObjectPermission.objects.filter(user=instance.user,object_id=project.id,content_type=ContentType.objects.get_for_model(Project)).count() == 0:
                    ObjectPermission.objects.create(content_object=project, user=instance.user,permission_type=ObjectPermission.PERMISSION_VIEW)


@receiver(post_save, sender=ObjectPermission)
def delete_permission_none(sender, instance, **kwargs):
    if instance.permission_type == instance.PERMISSION_NONE:
        instance.delete()


@receiver(pre_save, sender=ObjectPermission)
def delete_existing_permission(sender, instance, **kwargs):
    ObjectPermission.objects.filter(
        user=instance.user,
        content_type=instance.content_type,
        object_id=instance.object_id,
    ).delete()


@receiver(pre_save, sender=LiveProjectUser)
def delete_existing_role(sender, instance, **kwargs):
    LiveProjectUser.objects.filter(
        user=instance.user,
        liveproject=instance.liveproject,
    ).delete()


@receiver(post_save, sender=LiveProjectUser)
def delete_role_none(sender, instance, **kwargs):
    if instance.role_type == instance.ROLE_NONE:
        instance.delete()


@receiver(pre_delete, sender=ObjectPermission)
def remove_permissions_to_project_objects(sender, instance, **kwargs):
    if instance.content_type == ContentType.objects.get_for_model(Project):
        for workflow in instance.content_object.workflows.all():
            ObjectPermission.objects.filter(
                user=instance.user,
                content_type=ContentType.objects.get_for_model(workflow),
                object_id=workflow.get_subclass().id,
            ).delete()


@receiver(post_save, sender=Project)
def set_publication_of_project_objects(sender, instance, created, **kwargs):
    for workflow in instance.workflows.all():
        workflow.published = instance.published
        workflow.disciplines.set(instance.disciplines.all())
        workflow.save()


#    for outcome in instance.outcomes.all():
#        outcome.published = instance.published
#        outcome.disciplines.set(instance.disciplines.all())
#        outcome.save()


@receiver(post_save, sender=NodeWeek)
def set_node_type_default(sender, instance, created, **kwargs):
    node = instance.node
    try:
        node.node_type = instance.week.week_type
        node.save()
    except ValidationError:
        print("couldn't set default node type")


@receiver(post_save, sender=WeekWorkflow)
def set_week_type_default(sender, instance, created, **kwargs):
    week = instance.week
    try:
        week.week_type = instance.workflow.get_subclass().WORKFLOW_TYPE
        week.save()
    except ValidationError:
        print("couldn't set default week type")


@receiver(post_save, sender=OutcomeOutcome)
def set_outcome_depth_default(sender, instance, created, **kwargs):
    if created:
        try:
            set_list = list(instance.parent.sets.all())
            outcomes, outcomeoutcomes = get_all_outcomes_for_outcome(
                instance.child
            )
            outcomenodes_to_add = OutcomeNode.objects.filter(
                outcome=instance.parent
            )
            horizontallinks_to_add = OutcomeHorizontalLink.objects.filter(
                parent_outcome=instance.parent
            )
            for outcomeoutcome in [instance] + list(outcomeoutcomes):
                child = outcomeoutcome.child
                parent = outcomeoutcome.parent
                child.depth = parent.depth + 1
                child.sets.clear()
                child.sets.add(*set_list)
                child.save()
                for outcomenode in outcomenodes_to_add:
                    OutcomeNode.objects.create(
                        node=outcomenode.node,
                        outcome=outcomeoutcome.child,
                        degree=outcomenode.degree,
                    )
                for horizontallink in horizontallinks_to_add:
                    OutcomeHorizontalLink.objects.create(
                        outcome=horizontallink.outcome,
                        parent_outcome=outcomeoutcome.child,
                        degree=horizontallink.degree,
                    )
        except ValidationError:
            print("couldn't set default outcome depth or copy sets")


@receiver(post_save, sender=Node)
def create_default_node_content(sender, instance, created, **kwargs):
    if created and instance.is_original:
        # If this is an activity-level node, set the autolinks to true
        if instance.node_type == instance.ACTIVITY_NODE:
            instance.has_autolink = True
            instance.save()
        elif instance.node_type == instance.PROGRAM_NODE:
            instance.time_units = instance.CREDITS
            instance.save()


@receiver(post_save, sender=Activity)
def create_default_activity_content(sender, instance, created, **kwargs):
    if created and instance.is_original:
        # If the activity is newly created, add the default columns
        cols = instance.DEFAULT_COLUMNS
        for i, col in enumerate(cols):
            instance.columns.create(
                through_defaults={"rank": i},
                column_type=col,
                author=instance.author,
            )

        instance.weeks.create(
            week_type=Week.PART,
            author=instance.author,
            is_strategy=instance.is_strategy,
        )
        instance.save()


@receiver(post_save, sender=Course)
def create_default_course_content(sender, instance, created, **kwargs):
    if created and instance.is_original:
        # If the activity is newly created, add the default columns
        cols = instance.DEFAULT_COLUMNS
        for i, col in enumerate(cols):
            instance.columns.create(
                through_defaults={"rank": i},
                column_type=col,
                author=instance.author,
            )

        instance.weeks.create(
            week_type=Week.WEEK,
            author=instance.author,
            is_strategy=instance.is_strategy,
        )
        instance.time_units = instance.CREDITS
        instance.save()


@receiver(post_save, sender=Program)
def create_default_program_content(sender, instance, created, **kwargs):
    if created and instance.is_original:
        # If the activity is newly created, add the default columns
        cols = instance.DEFAULT_COLUMNS
        for i, col in enumerate(cols):
            instance.columns.create(
                through_defaults={"rank": i},
                column_type=col,
                author=instance.author,
            )

        instance.weeks.create(
            week_type=Week.TERM,
            author=instance.author,
            is_strategy=instance.is_strategy,
        )
        instance.condensed = True
        instance.save()


@receiver(post_save, sender=Project)
@receiver(post_save, sender=Workflow)
def add_default_editor_workflow(sender, instance, created, **kwargs):
    if created and instance.author is not None:
        ObjectPermission.objects.create(
            content_object=instance,
            user=instance.author,
            permission_type=ObjectPermission.PERMISSION_EDIT,
        )


@receiver(post_save, sender=Activity)
@receiver(post_save, sender=Course)
@receiver(post_save, sender=Program)
def add_default_editor_other_workflow(sender, instance, created, **kwargs):
    instance = Workflow.objects.get(pk=instance.pk)
    if created and instance.author is not None:
        ObjectPermission.objects.create(
            content_object=instance,
            user=instance.author,
            permission_type=ObjectPermission.PERMISSION_EDIT,
        )


@receiver(post_save, sender=WorkflowProject)
def set_publication_workflow(sender, instance, created, **kwargs):
    if created:
        # Set the workflow's publication status to that of the project
        workflow = instance.workflow
        workflow.published = instance.project.published
        workflow.disciplines.set(instance.project.disciplines.all())
        if instance.project.author != workflow.get_subclass().author:
            ObjectPermission.objects.create(
                content_object=workflow.get_subclass(),
                user=instance.project.author,
                permission_type=ObjectPermission.PERMISSION_EDIT,
            )
        for op in ObjectPermission.objects.filter(
            content_type=ContentType.objects.get_for_model(instance.project),
            object_id=instance.project.id,
        ):
            ObjectPermission.objects.create(
                content_object=workflow,
                user=op.user,
                permission_type=op.permission_type,
            )
        workflow.save()


# @receiver(post_save, sender=OutcomeProject)
# def set_publication_outcome(sender, instance, created, **kwargs):
#    if created:
#        # Set the workflow's publication status to that of the project
#        outcome = instance.outcome
#        outcome.published = instance.project.published
#        outcome.disciplines.set(instance.project.disciplines.all())
#        if instance.project.author != outcome.author:
#            ObjectPermission.objects.create(
#                content_object=outcome,
#                user=instance.project.author,
#                permission_type=ObjectPermission.PERMISSION_EDIT,
#            )
#        for op in ObjectPermission.objects.filter(
#            content_type=ContentType.objects.get_for_model(instance.project),
#            object_id=instance.project.id,
#        ):
#            ObjectPermission.objects.create(
#                content_object=outcome,
#                user=op.user,
#                permission_type=op.permission_type,
#            )
#        outcome.save()


@receiver(post_save, sender=WeekWorkflow)
def switch_week_to_static(sender, instance, created, **kwargs):
    if created:
        if instance.workflow.static:
            for node in instance.week.nodes.all():
                node.students.add(*list(instance.workflow.students.all()))


# Live project receivers


@receiver(pre_save, sender=LiveProject)
def add_all_workflows(sender, instance, **kwargs):
    if instance.default_all_workflows_visible:
        if LiveProject.objects.get(
            pk=instance.pk
        ).default_all_workflows_visible:
            return
        instance.visible_workflows.add(
            *Workflow.objects.filter(project=instance.project, deleted=False)
        )


@receiver(post_save, sender=LiveProject)
def add_owner_to_liveproject(sender, instance, created, **kwargs):
    if created:
        LiveProjectUser.objects.create(
            liveproject=instance,
            user=instance.project.author,
            role_type=LiveProjectUser.ROLE_TEACHER,
        )


@receiver(post_save, sender=WorkflowProject)
def add_to_visible_workflows(sender, instance, created, **kwargs):
    if created and abs(
        instance.workflow.created_on - timezone.now()
    ) < timezone.timedelta(seconds=10):
        project = instance.project
        if project is not None:
            try:
                liveproject = project.liveproject
                if liveproject.default_all_workflows_visible:
                    liveproject.visible_workflows.add(instance.workflow)
            except AttributeError:
                pass


@receiver(pre_save, sender=Workflow)
def add_or_remove_visible_workflow_on_delete_restore(
    sender, instance, **kwargs
):
    workflow = Workflow.objects.get(pk=instance.pk)
    if workflow.deleted != instance.deleted:
        project = instance.get_project()
        if project is not None:
            try:
                liveproject = project.liveproject
                if instance.deleted:
                    liveproject.visible_workflows.remove(workflow)
                elif liveproject.default_all_workflows_visible:
                    liveproject.visible_workflows.add(workflow)
            except AttributeError:
                pass


@receiver(post_save, sender=LiveProjectUser)
def add_user_to_assignments(sender, instance, created, **kwargs):
    if (
        instance.role_type == LiveProjectUser.ROLE_STUDENT
        and instance.liveproject.default_assign_to_all
    ):
        for assignment in LiveAssignment.objects.filter(
            liveproject=instance.liveproject,
        ).exclude(userassignment__user=instance.user):
            UserAssignment.objects.create(
                user=instance.user, assignment=assignment
            )


@receiver(post_save, sender=LiveProjectUser)
def add_user_to_assignments(sender, instance, created, **kwargs):
    if instance.role_type == LiveProjectUser.ROLE_NONE:
        UserAssignment.objects.filter(
            assignment__liveproject=instance.liveproject
        ).delete()


@receiver(post_save, sender=LiveAssignment)
def live_assignment_creation_defaults(sender, instance, created, **kwargs):
    if created:
        liveproject = instance.liveproject
        instance.self_reporting = liveproject.default_self_reporting
        instance.single_completion = liveproject.default_single_completion
        if liveproject.default_assign_to_all:
            students = LiveProjectUser.objects.filter(
                liveproject=liveproject, role_type=LiveProjectUser.ROLE_STUDENT
            )
            for student in students:
                UserAssignment.objects.create(
                    user=student.user, assignment=instance
                )


